import json
from pathlib import Path

import pytest

import bluepysnap.config as test_module
from bluepysnap.exceptions import BluepySnapError

from utils import TEST_DATA_DIR, copy_config, edit_config


def parse(path):
    with open(path) as fd:
        config = json.load(fd)

    return test_module.Parser.parse(config, str(Path(path).parent))


def test_parse():
    actual = parse(str(TEST_DATA_DIR / "circuit_config.json"))

    # check double resolution and '.' works: $COMPONENT_DIR -> $BASE_DIR -> '.'
    assert actual["components"]["morphologies_dir"] == str(TEST_DATA_DIR / "morphologies")

    # check resolution and './' works: $NETWORK_DIR -> './'
    assert actual["networks"]["nodes"][0]["nodes_file"] == str(TEST_DATA_DIR / "nodes.h5")

    # check resolution of '../' works: $PARENT --> '../'
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["manifest"]["$PARENT"] = "../"
            config["components"]["other"] = "$PARENT/other"

        actual = parse(config_path)
        assert actual["components"]["other"] == str(Path(config_path.parent / "../other").resolve())

    # check resolution of '../' works in a path outside manifest
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["components"]["other"] = "../other"

        actual = parse(config_path)
        assert actual["components"]["other"] == str(Path(config_path.parent / "../other").resolve())

    # check resolution without manifest of '../' works in a path outside
    # i.e. : self.manifest contains the configdir even if manifest is not here
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            for k in list(config):
                config.pop(k)
            config["something"] = "../other"
        actual = parse(config_path)
        assert actual["something"] == str(Path(config_path.parent / "../other").resolve())

    # check resolution with multiple slashes
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["something"] = "$COMPONENT_DIR/something////else"
        actual = parse(config_path)
        assert actual["something"] == str(Path(config_path.parent) / "something" / "else")

    # check resolution with $ in a middle of the words
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["something"] = "$COMPONENT_DIR/somet$hing/else"
        actual = parse(config_path)
        assert actual["something"] == str(Path(config_path.parent) / "somet$hing" / "else")

    # check resolution with relative path without "." in the manifest
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["manifest"]["$NOPOINT"] = "nopoint"
            config["components"]["other"] = "$NOPOINT/other"
        actual = parse(config_path)
        assert actual["components"]["other"] == str(Path(config_path.parent) / "nopoint" / "other")

    # check resolution for non path objects
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            for k in list(config):
                config.pop(k)
            config["string"] = "string"
            config["int"] = 1
            config["double"] = 0.2
            # just to check because we use starting with '.' as a special case
            config["tricky_double"] = 0.2
            config["path"] = "./path"

        actual = parse(config_path)

        # string
        assert actual["string"] == "string"
        # int
        assert actual["int"] == 1
        # double
        assert actual["double"] == 0.2
        assert actual["tricky_double"] == 0.2
        # path
        assert actual["path"] == str(Path(config_path.parent / "./path").resolve())


def test_bad_manifest():
    # 2 anchors would result in the absolute path of the last one : misleading
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["manifest"]["$COMPONENT_DIR"] = "$BASE_DIR/$NETWORK_DIR"
        with pytest.raises(BluepySnapError):
            parse(config_path)

    # same but not in the manifest
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["components"]["other"] = "$COMPONENT_DIR/$BASE_DIR"
        with pytest.raises(BluepySnapError):
            parse(config_path)

    # manifest value not a string
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["manifest"]["$COMPONENT_DIR"] = 42
        with pytest.raises(BluepySnapError):
            parse(config_path)

    # relative path with an anchor in the middle is not allowed this breaks the purpose of the
    # anchors (they are not just generic placeholders)
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["components"]["other"] = "something/$COMPONENT_DIR/"
        with pytest.raises(BluepySnapError):
            parse(config_path)

    # abs path with an anchor in the middle is not allowed
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["components"]["other"] = "/something/$COMPONENT_DIR/"
        with pytest.raises(BluepySnapError):
            parse(config_path)

    # unknown anchor
    with copy_config() as config_path:
        with edit_config(config_path) as config:
            config["components"]["other"] = "$UNKNOWN/something/"
        with pytest.raises(KeyError):
            parse(config_path)


def test_simulation_config():
    actual = test_module.SimulationConfig.from_config(
        str(TEST_DATA_DIR / "simulation_config.json")
    ).to_dict()
    assert actual["target_simulator"] == "CORENEURON"
    assert actual["network"] == str(Path(TEST_DATA_DIR / "circuit_config.json").resolve())
    assert actual["mechanisms_dir"] == str(
        Path(TEST_DATA_DIR / "../shared_components_mechanisms").resolve()
    )
    assert actual["conditions"]["celsius"] == 34.0
    assert actual["conditions"]["v_init"] == -80


def test__resolve_population_configs():
    node = {
        "nodes_file": "nodes_path",
        "populations": {"node_pop": {"changed_node": "changed", "added_this": "node_test"}},
    }

    edge = {
        "edges_file": "edges_path",
        "populations": {"edge_pop": {"changed_edge": "changed", "added_this": "edge_test"}},
    }

    config = {
        "components": {
            "unchanged": True,
            "changed_node": "unchanged",
            "changed_edge": "unchanged",
        },
        "networks": {
            "nodes": [node],
            "edges": [edge],
        },
    }
    expected_node_pop = {
        "unchanged": True,
        "changed_node": "changed",
        "changed_edge": "unchanged",
        "added_this": "node_test",
        "nodes_file": "nodes_path",
    }
    expected_edge_pop = {
        "unchanged": True,
        "changed_node": "unchanged",
        "changed_edge": "changed",
        "added_this": "edge_test",
        "edges_file": "edges_path",
    }
    expected = {
        "nodes": {"node_pop": expected_node_pop},
        "edges": {"edge_pop": expected_edge_pop},
    }
    res = test_module.CircuitConfig._resolve_population_configs(config)

    assert res == expected



import pickle
from unittest.mock import PropertyMock, patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from numpy import dtype

import bluepysnap.nodes as test_module
from bluepysnap.circuit import Circuit
from bluepysnap.circuit_ids import CircuitNodeIds
from bluepysnap.circuit_ids_types import IDS_DTYPE, CircuitNodeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.node_sets import NodeSets

from utils import PICKLED_SIZE_ADJUSTMENT, TEST_DATA_DIR


class TestNodes:
    def setup_method(self):
        circuit = Circuit(str(TEST_DATA_DIR / "circuit_config.json"))
        self.test_obj = test_module.Nodes(circuit)

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.NodePopulation)
        with pytest.raises(BluepySnapError):
            self.test_obj["unknown"]

    def test_iter(self):
        assert sorted(self.test_obj) == ["default", "default2"]

    def test_population_names(self):
        assert self.test_obj.population_names == ["default", "default2"]

    def test_keys_names(self):
        assert list(self.test_obj.keys()) == ["default", "default2"]

    def test_values_population(self):
        values = list(self.test_obj.values())
        assert isinstance(values[0], test_module.NodePopulation)
        assert values[0].name == "default"

        assert isinstance(values[1], test_module.NodePopulation)
        assert values[1].name == "default2"

    def test_items(self):
        keys, values = zip(*self.test_obj.items())
        assert keys == ("default", "default2")
        assert isinstance(values[0], test_module.NodePopulation)
        assert values[0].name == "default"

        assert isinstance(values[1], test_module.NodePopulation)
        assert values[1].name == "default2"

    def test_size(self):
        assert self.test_obj.size == 7

    def test_property_names(self):
        assert self.test_obj.property_names == {
            "rotation_angle_zaxis",
            "y",
            "layer",
            "mtype",
            "model_type",
            "z",
            "x",
            "rotation_angle_yaxis",
            "morphology",
            "rotation_angle_xaxis",
            "model_template",
            "other1",
            "other2",
            "@dynamics:holding_current",
        }

    def test_property_value(self):
        assert self.test_obj.property_values("mtype") == {"L2_X", "L7_X", "L9_Z", "L8_Y", "L6_Y"}
        assert self.test_obj.property_values("other2") == {10, 11, 12, 13}

    def test_ids(self):
        np.random.seed(0)

        # None --> CircuitNodeIds with all ids
        tested = self.test_obj.ids()
        expected = CircuitNodeIds.from_dict({"default": [0, 1, 2], "default2": [0, 1, 2, 3]})
        assert tested == expected
        npt.assert_equal(tested.get_ids().dtype, IDS_DTYPE)

        # CircuitNodeIds --> CircuitNodeIds and check if the population and node ids exist
        ids = CircuitNodeIds.from_arrays(["default", "default2"], [0, 3])
        tested = self.test_obj.ids(ids)
        assert tested == ids

        # default3 population does not exist and is asked explicitly
        with pytest.raises(BluepySnapError):
            ids = CircuitNodeIds.from_arrays(["default", "default3"], [0, 3])
            self.test_obj.ids(ids)

        # (default2, 5) does not exist and is asked explicitly
        with pytest.raises(BluepySnapError):
            ids = CircuitNodeIds.from_arrays(["default", "default2"], [0, 5])
            self.test_obj.ids(ids)

        # single node ID --> CircuitNodeIds return populations with the 0 id
        expected = CircuitNodeIds.from_arrays(["default", "default2"], [0, 0])
        tested = self.test_obj.ids(0)
        assert tested == expected

        # single node ID --> CircuitNodeIds raise if the ID is not in all population
        with pytest.raises(BluepySnapError):
            self.test_obj.ids(3)

        # seq of node ID --> CircuitNodeIds return populations with the array of ids
        expected = CircuitNodeIds.from_arrays(
            ["default", "default", "default2", "default2"], [0, 1, 0, 1]
        )
        tested = self.test_obj.ids([0, 1])
        assert tested == expected
        tested = self.test_obj.ids((0, 1))
        assert tested == expected
        tested = self.test_obj.ids(np.array([0, 1]))
        assert tested == expected

        ids = [CircuitNodeId("default", 0), CircuitNodeId("default", 1)]
        assert self.test_obj.ids(ids) == CircuitNodeIds.from_dict({"default": [0, 1]})

        with pytest.raises(BluepySnapError):
            ids = [CircuitNodeId("default", 0), ("default", 1)]
            self.test_obj.ids(ids)

        # seq node ID --> CircuitNodeIds raise if on ID is not in all populations
        with pytest.raises(BluepySnapError):
            self.test_obj.ids([0, 1, 2, 3])

        # node sets
        assert self.test_obj.ids("Layer2") == CircuitNodeIds.from_arrays(
            ["default", "default2"], [0, 3]
        )
        assert self.test_obj.ids("Layer23") == CircuitNodeIds.from_arrays(
            ["default", "default2"], [0, 3]
        )
        assert self.test_obj.ids("Population_default_L6_Y_Node2") == CircuitNodeIds.from_arrays(
            ["default"], [2]
        )
        assert self.test_obj.ids(
            "combined_combined_Node0_L6_Y__Node12_L6_Y__"
        ) == CircuitNodeIds.from_arrays(["default", "default", "default", "default2"], [0, 1, 2, 3])

        # Mapping --> CircuitNodeIds query on the populations empty dict return all
        assert self.test_obj.ids({}) == self.test_obj.ids()

        assert self.test_obj.ids({"$and": []}) == self.test_obj.ids()
        assert self.test_obj.ids({"$or": []}) == CircuitNodeIds.from_dict({})

        # Mapping --> CircuitNodeIds query on the populations
        tested = self.test_obj.ids({"layer": 2})
        expected = CircuitNodeIds.from_arrays(["default", "default2"], [0, 3])
        assert tested == expected

        # Mapping --> CircuitNodeIds query on the populations no raise if not in one of the pop
        tested = self.test_obj.ids({"other1": ["A", "D"]})
        expected = CircuitNodeIds.from_arrays(["default2", "default2"], [0, 3])
        assert tested == expected

        # Mapping --> CircuitNodeIds query on the populations no raise if not in one of the pop
        tested = self.test_obj.ids({"other1": ["A", "D"], "layer": 2})
        expected = CircuitNodeIds.from_arrays(["default2"], [3])
        assert tested == expected

        # Mapping --> CircuitNodeIds query on the populations no raise if not in one of the pop
        tested = self.test_obj.ids({"$or": [{"other1": ["A", "D"]}, {"layer": 2}]})
        expected = CircuitNodeIds.from_arrays(["default", "default2", "default2"], [0, 0, 3])
        assert tested == expected

        # Mapping --> CircuitNodeIds query on the populations no raise if not in one of the pop
        tested = self.test_obj.ids({"$and": [{"other1": ["A", "D"]}, {"layer": 2}]})
        expected = CircuitNodeIds.from_arrays(["default2"], [3])
        assert tested == expected

        # Mapping --> CircuitNodeIds query on the population node ids with mapping.
        # single pop
        tested = self.test_obj.ids({"population": "default"})
        expected = self.test_obj.ids().filter_population("default")
        assert tested == expected
        # multiple pop
        tested = self.test_obj.ids({"population": ["default", "default2"]})
        expected = self.test_obj.ids()
        assert tested == expected
        # not existing pop (should not raise)
        tested = self.test_obj.ids({"population": "default4"})
        expected = CircuitNodeIds.from_arrays([], [])
        assert tested == expected

        # single pop and node ids
        tested = self.test_obj.ids({"population": ["default"], "node_id": [1, 2]})
        expected = CircuitNodeIds.from_arrays(["default", "default"], [1, 2])
        assert tested == expected
        # single pop and node ids with not present node id (should not raise)
        tested = self.test_obj.ids({"population": ["default"], "node_id": [1, 5]})
        expected = CircuitNodeIds.from_arrays(["default"], [1])
        assert tested == expected
        # not existing node ids (should not raise)
        tested = self.test_obj.ids({"population": ["default"], "node_id": [5, 6, 7]})
        expected = CircuitNodeIds.from_arrays([], [])
        assert tested == expected

        # multiple pop and node ids
        tested = self.test_obj.ids({"population": ["default", "default2"], "node_id": [1, 0]})
        expected = CircuitNodeIds.from_arrays(
            ["default", "default", "default2", "default2"], [1, 0, 1, 0]
        )
        assert tested == expected
        # multiple pop and node ids with not present node id (should not raise)
        tested = self.test_obj.ids({"population": ["default", "default2"], "node_id": [1, 0, 3]})
        expected = CircuitNodeIds.from_arrays(
            ["default", "default", "default2", "default2", "default2"], [1, 0, 1, 0, 3]
        )
        assert tested == expected

        # Check operations on global ids
        ids = self.test_obj.ids()
        assert ids.filter_population("default").append(ids.filter_population("default2")) == ids

        expected = CircuitNodeIds.from_arrays(["default2", "default2"], [0, 1])
        assert ids.filter_population("default2").limit(2) == expected

        tested = self.test_obj.ids(sample=2)
        expected = CircuitNodeIds.from_arrays(["default2", "default"], [3, 2], sort_index=False)
        assert tested == expected

        tested = self.test_obj.ids(limit=4)
        expected = CircuitNodeIds.from_dict({"default": [0, 1, 2], "default2": [0]})
        assert tested == expected

    def test_get(self):
        # return all properties for all the ids
        tested = self.test_obj.get()
        tested = pd.concat(df for _, df in tested)
        assert tested.shape == (self.test_obj.size, len(self.test_obj.property_names))

        # put NaN for the undefined values : only values for default2 in dropna
        assert len(tested.dropna()) == 4

        # the index of the dataframe is the index from all the NodeCircuitIds
        pdt.assert_index_equal(tested.index, self.test_obj.ids().index)

        # tested accessing data via circuit ids
        tested_ids = self.test_obj.ids({"population": "default"})
        assert tested.loc[tested_ids, "layer"].tolist() == [2, 6, 6]

        # tested columns
        tested = self.test_obj.get(properties=["other2", "other1", "layer"])
        tested = pd.concat(df for _, df in tested)
        expected = pd.DataFrame(
            {
                "other2": np.array([np.NaN, np.NaN, np.NaN, 10, 11, 12, 13], dtype=float),
                "other1": np.array([np.NaN, np.NaN, np.NaN, "A", "B", "C", "D"], dtype=object),
                "layer": np.array([2, 6, 6, 7, 8, 8, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "node_ids"],
            ),
        )
        pdt.assert_frame_equal(tested[expected.columns], expected)

        tested = self.test_obj.get(
            group={"population": "default2"}, properties=["other2", "other1", "layer"]
        )
        expected = pd.DataFrame(
            {
                "other2": np.array([10, 11, 12, 13], dtype=int),
                "other1": np.array(["A", "B", "C", "D"], dtype=object),
                "layer": np.array([7, 8, 8, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "node_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        with pytest.raises(KeyError, match="'default'"):
            tested.loc[("default", 0)]

        tested = self.test_obj.get(
            group={"population": "default"}, properties=["other2", "other1", "layer"]
        )
        expected = pd.DataFrame(
            {
                "layer": np.array([2, 6, 6], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                ],
                names=["population", "node_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        tested = self.test_obj.get(properties="layer")
        expected = pd.DataFrame(
            {
                "layer": np.array([2, 6, 6, 7, 8, 8, 2], dtype=int),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "node_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        tested = self.test_obj.get(properties="other2")
        expected = pd.DataFrame(
            {
                "other2": np.array([np.NaN, np.NaN, np.NaN, 10, 11, 12, 13], dtype=float),
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ("default", 0),
                    ("default", 1),
                    ("default", 2),
                    ("default2", 0),
                    ("default2", 1),
                    ("default2", 2),
                    ("default2", 3),
                ],
                names=["population", "node_ids"],
            ),
        )
        tested = pd.concat(df for _, df in tested)
        pdt.assert_frame_equal(tested, expected)

        with pytest.raises(BluepySnapError, match="Unknown properties required: {'unknown'}"):
            next(self.test_obj.get(properties=["other2", "unknown"]))

        with pytest.raises(BluepySnapError, match="Unknown properties required: {'unknown'}"):
            next(self.test_obj.get(properties="unknown"))

    def test_functionality_with_separate_node_set(self):
        with pytest.raises(BluepySnapError, match="Undefined node set"):
            self.test_obj.ids("ExtraLayer2")

        node_sets = NodeSets.from_file(str(TEST_DATA_DIR / "node_sets_extra.json"))

        assert self.test_obj.ids(node_sets["ExtraLayer2"]) == CircuitNodeIds.from_arrays(
            ["default", "default2"], [0, 3]
        )

        with pytest.raises(BluepySnapError, match="Undefined node set"):
            next(self.test_obj.get("ExtraLayer2"))

        tested = pd.concat(df for _, df in self.test_obj.get(node_sets["ExtraLayer2"]))
        expected = pd.concat(df for _, df in self.test_obj.get("Layer2"))
        pdt.assert_frame_equal(tested, expected)

    def test_pickle(self, tmp_path):
        pickle_path = tmp_path / "pickle.pkl"

        # trigger some cached properties, to makes sure they aren't being pickeld
        self.test_obj.size
        self.test_obj.property_names

        with open(pickle_path, "wb") as fd:
            pickle.dump(self.test_obj, fd)

        with open(pickle_path, "rb") as fd:
            test_obj = pickle.load(fd)

        assert pickle_path.stat().st_size < 100 + PICKLED_SIZE_ADJUSTMENT
        assert test_obj.size == 7



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""SONATA network config parsing."""

import copy
import json
from collections.abc import Iterable, Mapping
from pathlib import Path

import libsonata
from libsonata import CircuitConfigStatus

from bluepysnap.exceptions import BluepySnapError

# List of keys which are expected to have paths
EXPECTED_PATH_KEYS = {
    "morphologies_dir",
    "biophysical_neuron_models_dir",
    "vasculature_file",
    "vasculature_mesh",
    "endfeet_meshes_file",
    "microdomains_file",
    "neurolucida-asc",
    "h5v1",
    "edges_file",
    "nodes_file",
    "edges_type_file",
    "nodes_type_file",
    "node_sets_file",
    "output_dir",
    "network",
    "mechanisms_dir",
}


class Parser:
    """SONATA network config parser.

    See Also:
        https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md#network_config
    """

    def __init__(self, config, config_dir):
        """Initializes a Resolver object.

        Args:
            config (dict): Dict containing the config.
            config_dir(str):  Path to the directory containing the config file.

        Returns:
             Parser: A Parser object.
        """
        content = config.copy()
        manifest = content.pop("manifest", None) or {}
        self.manifest = Parser._resolve_manifest(manifest, config_dir)
        self.content = content

    @staticmethod
    def _resolve_manifest(manifest, configdir):
        result = manifest.copy()

        for k, v in result.items():
            if not isinstance(v, str):
                raise BluepySnapError(f"{v} should be a string value.")
            if not Path(v).is_absolute() and not v.startswith("$"):
                result[k] = str(Path(configdir, v).resolve())

        while True:
            update = False
            for k, v in result.items():
                if v.count("$") > 1:
                    raise BluepySnapError(
                        f"{k} is not a valid anchor : contains more than one sub anchor."
                    )
                if v.startswith("$"):
                    tokens = v.split("/", 1)
                    resolved = result[tokens[0]]
                    if "$" not in resolved:
                        result[k] = str(Path(resolved, *tokens[1:]))
                        update = True
            if not update:
                break

        assert "${configdir}" not in result
        result["${configdir}"] = configdir

        return result

    def _resolve_string(self, value, key):
        # not a startswith to detect the badly placed anchors
        if "$" in value:
            vs = [self.manifest[v] if v.startswith("$") else v for v in value.split("/")]
            abs_paths = [v for v in vs[1:] if v.startswith("/")]
            if len(abs_paths) != 0:
                raise BluepySnapError(
                    f"Misplaced anchors in : {value}. Please verify your '$' usage."
                )
            return str(Path(*vs))
        # only way to know if value is a relative path or a normal string
        elif value.startswith(".") or key in EXPECTED_PATH_KEYS:
            return str(Path(self.manifest["${configdir}"], value).resolve())
        else:
            # we cannot know if a string is a path or not if it does not contain anchor or .
            return value

    def _resolve(self, value, key=None):
        if isinstance(value, Mapping):
            return {k: self._resolve(v, k) for k, v in value.items()}
        elif isinstance(value, str):
            return self._resolve_string(value, key)
        elif isinstance(value, Iterable):
            return [self._resolve(v) for v in value]
        else:
            return value

    def resolve(self):
        """Resolve variables in config file paths."""
        return self._resolve(self.content)

    @staticmethod
    def parse(config, configdir):
        """Parse SONATA network config."""
        return Parser(config, configdir).resolve()


class Config:
    """Common config class."""

    def __init__(self, config, config_class):
        """Initializes the Config class.

        Args:
            config (str): Path to the configuration file
            config_class (class): libsonata class corresponding to the configuration file, either
                libsonata.CircuitConfig or libsonata.SimulationConfig
        """
        self._config_dir = str(Path(config).parent.absolute())
        self._libsonata = config_class.from_file(config)

    @property
    def to_libsonata(self):
        """Return the libsonata instance of the config."""
        return self._libsonata

    def to_dict(self):
        """Return the configuration as a dict with absolute paths."""
        return Parser.parse(json.loads(self._libsonata.expanded_json), self._config_dir)


class CircuitConfig(Config):
    """Handle CircuitConfig."""

    def __init__(self, *args):
        """Initializes circuit config."""
        super().__init__(*args)
        self._populations = self._resolve_population_configs(self.to_dict())

    @classmethod
    def from_config(cls, config_path):
        """Instantiate the config class from circuit configuration."""
        return cls(config_path, libsonata.CircuitConfig)

    @property
    def node_populations(self):
        """Access node population configs."""
        return self._populations["nodes"]

    @property
    def edge_populations(self):
        """Access edge population configs."""
        return self._populations["edges"]

    @property
    def status(self) -> CircuitConfigStatus:
        """Return status of the config."""
        return self._libsonata.config_status

    @staticmethod
    def _resolve_population_configs(config):
        """Resolves population configs for the node and edge populations."""
        networks = config.get("networks") or {}
        components = config.get("components") or {}

        def resolve_network_populations(element_type):
            populations = {}
            type_file_key = f"{element_type}_file"
            element_dicts = networks.get(element_type) or []

            for elem_dict in element_dicts:
                elem_pops = elem_dict.get("populations") or {}

                for pop, cfg in elem_pops.items():
                    populations[pop] = copy.deepcopy(components)
                    populations[pop].update(cfg or {})

                    # add h5 filepath for simpler access
                    if type_file_key in elem_dict:
                        populations[pop][type_file_key] = elem_dict[type_file_key]

            return populations

        return {
            "nodes": resolve_network_populations("nodes"),
            "edges": resolve_network_populations("edges"),
        }


class SimulationConfig(Config):
    """Handle SimulationConfig."""

    @classmethod
    def from_config(cls, config_path):
        """Instantiate the config class from simulation configuration."""
        return cls(config_path, libsonata.SimulationConfig)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Exceptions used throughout the library."""


class BluepySnapError(Exception):
    """Base SNAP exception."""


class BluepySnapDeprecationError(Exception):
    """SNAP deprecation exception."""


class BluepySnapDeprecationWarning(DeprecationWarning):
    """SNAP deprecation warning."""


class BluepySnapValidationError:
    """Error used for reporting of validation errors."""

    FATAL = "FATAL"
    WARNING = "WARNING"
    INFO = "INFO"

    def __init__(self, level, message=None):
        """Error.

        Args:
            level (str): error level
            message (str|None): message
        """
        self.level = level
        self.message = message

    def __str__(self):
        """Returns only message by default."""
        return str(self.message)

    __repr__ = __str__

    def __eq__(self, other):
        """Two errors are equal if inherit from Error and their level, message are equal."""
        if not isinstance(other, BluepySnapValidationError):
            return False
        return self.level == other.level and self.message == other.message

    def __hash__(self):
        """Hash. Errors with the same level and message give the same hash."""
        return hash(self.level) ^ hash(self.message)

    @classmethod
    def warning(cls, message):
        """Shortcut for a warning.

        Args:
            message (str): text message

        Returns:
            Error: Error with level WARNING
        """
        return cls(cls.WARNING, message)

    @classmethod
    def fatal(cls, message):
        """Shortcut for a fatal error.

        Args:
            message (str): text message

        Returns:
            Error: Error with level FATAL
        """
        return cls(cls.FATAL, message)



import pytest

from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import ConstContainer


class Container(ConstContainer):
    VAR1 = "var1"


class SubContainer(Container):
    VAR2 = "var2"


class SubSubContainer1(SubContainer):
    VAR3 = "var3"


class SubSubContainer2(SubContainer):
    VAR4 = "var4"


class ComplexContainer(SubSubContainer1, SubSubContainer2):
    VAR5 = "var5"


class FailingContainer(SubContainer, int):
    VAR6 = "var6"


def test_list_keys():
    assert SubSubContainer1.key_set() == {"VAR1", "VAR2", "VAR3"}
    assert SubSubContainer2.key_set() == {"VAR1", "VAR2", "VAR4"}
    assert ComplexContainer.key_set() == {"VAR1", "VAR2", "VAR3", "VAR4", "VAR5"}

    with pytest.raises(BluepySnapError):
        FailingContainer.key_set()


def test_get():
    assert SubSubContainer1.get("VAR1") == "var1"
    assert SubSubContainer1.get("VAR3") == "var3"
    with pytest.raises(BluepySnapError):
        SubSubContainer1.get("VAR4")



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Frame report access."""
import logging

import numpy as np
import pandas as pd
from cached_property import cached_property
from libsonata import ElementReportReader, SonataError

import bluepysnap._plotting
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.utils import ensure_ids

L = logging.getLogger(__name__)


def _collect_population_reports(frame_report, cls):
    return {
        population: cls(frame_report, population) for population in frame_report.population_names
    }


class PopulationFrameReport:
    """Access to PopulationFrameReport data."""

    def __init__(self, frame_report, population_name):
        """Initializes a PopulationFrameReport object from a FrameReport.

        Args:
            frame_report (FrameReport): FrameReport containing this frame report population.
            population_name (str): the population name corresponding to this report.

        Returns:
            PopulationFrameReport: A PopulationFrameReport object.
        """
        self.frame_report = frame_report
        self._frame_population = self._get_reader(frame_report, population_name)
        self._population_name = population_name

    @staticmethod
    def _get_reader(frame_report, population_name):
        """Access to the population compartment reader."""
        return ElementReportReader(frame_report.to_libsonata.file_name)[population_name]

    @property
    def name(self):
        """Access to the population name."""
        return self._population_name

    def _resolve(self, group):
        """Transform a group into ids array.

        Notes:
            The type of ids depends on the type of report and so this function needs to be
            implemented for all type of reports. It can return node or edge ids or something else.
        """
        raise NotImplementedError

    @staticmethod
    def _wrap_columns(columns):
        """Allows to change the columns names if needed."""
        return columns

    def get(self, group=None, t_start=None, t_stop=None, t_step=None):
        """Fetch data from the report.

        Args:
            group (None/int/list/np.array/dict): Get frames filtered by :ref:`Group Concept`.
            t_start (float): Include only frames occurring at or after this time.
            t_stop (float): Include only frames occurring at or before this time.
            t_step (float): Optional time step, useful to reduce the number of samples.
                It should be a multiple of the report time step dt, and it's equal to dt by default.
                If the given t_step isn't an exact multiple, it's rounded to the closer multiple.
                Only the samples at t = t0 + k * t_step, for k = 0, 1... are returned,
                where t0 is the first sample time >= t_start.

        Returns:
            pandas.DataFrame: frame as columns indexed by timestamps.
        """
        t_stride = round(t_step / self.frame_report.dt) if t_step is not None else 1
        if t_stride < 1:
            msg = f"Invalid {t_step=}. It should be None or a multiple of {self.frame_report.dt}."
            raise BluepySnapError(msg)
        ids = self._resolve(group).tolist()
        try:
            view = self._frame_population.get(
                node_ids=ids, tstart=t_start, tstop=t_stop, tstride=t_stride
            )
        except SonataError as e:
            raise BluepySnapError(e) from e

        # cell ids and section ids in the columns are enforced to be int64
        # to avoid issues with numpy automatic conversions and to ensure that
        # the results are the same regardless of the libsonata version [NSETM-1766]
        res = pd.DataFrame(
            data=view.data,
            columns=pd.MultiIndex.from_arrays(ensure_ids(view.ids).T),
            index=view.times,
        ).sort_index(axis=1)

        # rename from multi index to index cannot be achieved easily through df.rename
        res.columns = self._wrap_columns(res.columns)
        return res

    @cached_property
    def node_ids(self):
        """Returns the node ids present in the report.

        Returns:
            np.Array: Numpy array containing the node_ids included in the report
        """
        return np.sort(ensure_ids(self._frame_population.get_node_ids()))


class FilteredFrameReport:
    """Access to filtered FrameReport data."""

    def __init__(self, frame_report, group=None, t_start=None, t_stop=None):
        """Initialize a FilteredFrameReport.

        A FilteredFrameReport is a lazy and cached object which contains the filtered data
        from all the populations of a report.

        Args:
            frame_report (FrameReport): The FrameReport to filter.
            group (None/int/list/np.array/dict): Get frames filtered by :ref:`Group Concept`.
            t_start (float): Include only frames occurring at or after this time.
            t_stop (float): Include only frames occurring at or before this time.

        Returns:
            FilteredFrameReport: A FilteredFrameReport object.
        """
        self.frame_report = frame_report
        self.group = group
        self.t_start = t_start
        self.t_stop = t_stop

    @cached_property
    def report(self):
        """Access to the report data.

        Returns:
            pandas.DataFrame: A DataFrame containing the data from the report. Row's indices are the
            different timestamps and the column's MultiIndex are:

            - (population_name, node_id, compartment id) for the CompartmentReport
            - (population_name, node_id) for the SomaReport
        """
        dataframes = {}
        for population in self.frame_report.population_names:
            frames = self.frame_report[population]
            ids = frames.nodes.ids(group=self.group, raise_missing_property=False)
            df = frames.get(group=ids, t_start=self.t_start, t_stop=self.t_stop)
            dataframes[population] = df
        # optimize when there is at most one non-empty df: use copy=False, and no need to sort
        if sum(not df.empty for df in dataframes.values()) <= 1:
            return pd.concat(dataframes, axis=1, copy=False)
        # when concatenating multiple df, don't use copy=False because 2x slower (Pandas 2.0.2)
        result = pd.concat(dataframes, axis=1)
        return result.sort_index(axis=0).sort_index(axis=1)

    # pylint: disable=protected-access
    trace = bluepysnap._plotting.frame_trace


class FrameReport:
    """Access to FrameReport data."""

    def __init__(self, simulation, report_name):
        """Initializes a FrameReport object from a simulation object.

        Args:
            simulation (Simulation): Simulation containing this frame report.
            report_name (str): The name of this frame report.

        Returns:
            FrameReport: A FrameReport object.
        """
        self._simulation = simulation
        self.name = report_name

    @cached_property
    def _frame_reader(self):
        """Access to the compartment report reader."""
        return ElementReportReader(self.to_libsonata.file_name)

    @property
    def to_libsonata(self):
        """Access libsonata instance of the report."""
        return self.simulation.to_libsonata.report(self.name)

    @property
    def config(self):
        """Access the report config."""
        return self.simulation.config["reports"][self.name]

    @property
    def time_start(self):
        """Returns the starting time of the report."""
        return self.to_libsonata.start_time

    @property
    def time_stop(self):
        """Returns the stopping time of the report."""
        return self.to_libsonata.end_time

    @property
    def dt(self):
        """Returns the frequency of reporting in milliseconds."""
        dt = self.to_libsonata.dt
        if dt != self.simulation.dt:
            L.warning("dt from the report differs from the global simulation dt.")
        return dt

    @property
    def time_units(self):
        """Returns the data unit for this report."""
        units = {self._frame_reader[pop].time_units for pop in self.population_names}
        if len(units) > 1:
            raise BluepySnapError("Multiple time units found in the different populations.")
        return units.pop()

    @cached_property
    def data_units(self):
        """Returns the data unit for this report."""
        units = {self._frame_reader[pop].data_units for pop in self.population_names}
        if len(units) > 1:
            raise BluepySnapError("Multiple data units found in the different populations.")
        return units.pop()

    @property
    def node_set(self):
        """Returns the name of the node set for the report."""
        return self.to_libsonata.cells

    @property
    def simulation(self):
        """Return the Simulation object related to this frame report."""
        return self._simulation

    @cached_property
    def population_names(self):
        """Returns the population names included in this report."""
        return sorted(self._frame_reader.get_population_names())

    @cached_property
    def _population_report(self):
        """Collect the different PopulationFrameReport."""
        return _collect_population_reports(self, PopulationFrameReport)

    def __getitem__(self, population_name):
        """Access the PopulationFrameReports corresponding to the population 'population_name'."""
        return self._population_report[population_name]

    def __iter__(self):
        """Allows iteration over the different PopulationFrameReports."""
        return iter(self._population_report)

    def filter(self, group=None, t_start=None, t_stop=None):
        """Returns a FilteredFrameReport.

        A FilteredFrameReport is a lazy and cached object which contains the filtered data
        from all the populations of a report.

        Args:
            group (None/int/list/np.array/dict): Get frames filtered by :ref:`Group Concept`.
            t_start (float): Include only frames occurring at or after this time.
            t_stop (float): Include only frames occurring at or before this time.

        Returns:
            FilteredFrameReport: A FilteredFrameReport object.
        """
        return FilteredFrameReport(self, group, t_start, t_stop)


class PopulationCompartmentReport(PopulationFrameReport):
    """Access to PopulationCompartmentsReport data."""

    @cached_property
    def nodes(self):
        """Returns the NodePopulation corresponding to this report."""
        return self.frame_report.simulation.circuit.nodes[self._population_name]

    def _resolve(self, group):
        """Transform a group into a node_id array."""
        return self.nodes.ids(group=group)


class CompartmentReport(FrameReport):
    """Access to a CompartmentsReport data."""

    @cached_property
    def _population_report(self):
        """Collect the different PopulationCompartmentsReport."""
        return _collect_population_reports(self, PopulationCompartmentReport)


class PopulationSomaReport(PopulationCompartmentReport):
    """Access to PopulationSomaReport data."""

    @staticmethod
    def _wrap_columns(columns):
        """Transform pandas.MultiIndex into pandas.Index for the pandas.DataFrame columns.

        Notes:
            the libsonata.ElementsReader.get() returns tuple as columns for the data. For the
            soma reports it means: pandas.MultiIndex([(0, 0), (1, 0), ..., (last_node_id, 0)]).
            So we convert this into pandas.Index([0,1,..., last_node_id]).
        """
        return columns.levels[0]


class SomaReport(FrameReport):
    """Access to a SomaReport data."""

    @cached_property
    def _population_report(self):
        """Collect the different PopulationSomasReport."""
        return _collect_population_reports(self, PopulationSomaReport)



import os
from unittest.mock import patch

import bluepysnap.settings as test_module


def test_str2bool():
    assert test_module.str2bool("1")
    assert test_module.str2bool("y")
    assert test_module.str2bool("YES")
    assert not test_module.str2bool("0")
    assert not test_module.str2bool("n")
    assert not test_module.str2bool("No")
    assert not test_module.str2bool(None)


def test_STRICT_MODE():
    with patch.dict(os.environ, {"BLUESNAP_STRICT_MODE": "1"}):
        test_module.load_env()
    assert test_module.STRICT_MODE



# Copyright (c) 2020-2021, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Module containing the Abstract classes for the Network."""
import abc

import numpy as np
import pandas as pd
from cached_property import cached_property

from bluepysnap import utils
from bluepysnap.exceptions import BluepySnapError


class NetworkObject(abc.ABC):
    """Abstract class for the top level NetworkObjects accessor."""

    _population_class = None

    def __init__(self, circuit):
        """Initialize the top level NetworkObjects accessor."""
        self._circuit = circuit

    def _get_populations(self, cls):
        """Collects the different NetworkObjectPopulation and returns them as a dict."""
        return {name: cls(self._circuit, name) for name in self.population_names}

    @cached_property
    def _populations(self):
        """Cached population dictionary."""
        return self._get_populations(self._population_class)

    @property
    @abc.abstractmethod
    def population_names(self):
        """Should define all sorted NetworkObjects population names from the Circuit."""

    def keys(self):
        """Returns iterator on the NetworkObjectPopulation names.

        Made to simulate the behavior of a dict.keys().
        """
        return (name for name in self.population_names)

    def values(self):
        """Returns iterator on the NetworkObjectPopulations.

        Made to simulate the behavior of a dict.values().
        """
        return (self[name] for name in self.population_names)

    def items(self):
        """Returns iterator on the tuples (population name, NetworkObjectPopulations).

        Made to simulate the behavior of a dict.items().
        """
        return ((name, self[name]) for name in self.population_names)

    def __getitem__(self, population_name):
        """Access the NetworkObjectPopulation corresponding to the population 'population_name'."""
        try:
            return self._populations[population_name]
        except KeyError as e:
            raise BluepySnapError(f"{population_name} not a {self.__class__} population.") from e

    def __iter__(self):
        """Allows iteration over the different NetworkObjectPopulation."""
        return iter(self.keys())

    @cached_property
    def size(self):
        """Total number of NetworkObject inside the circuit."""
        return sum(pop.size for pop in self.values())

    @cached_property
    def property_names(self):
        """Returns all the NetworkObject properties present inside the circuit."""
        return set(prop for pop in self.values() for prop in pop.property_names)

    def _get_ids_from_pop(self, fun_to_apply, returned_ids_cls, sample=None, limit=None):
        """Get CircuitIds of class 'returned_ids_cls' for all populations using 'fun_to_apply'.

        Args:
            fun_to_apply (function): A function that returns the list of IDs for each population
                and the population containing these IDs.
            returned_ids_cls (CircuitNodeIds/CircuitEdgeIds): the class for the CircuitIds.
            sample (int): If specified, randomly choose ``sample`` number of
                IDs from the match result. If the size of the sample is greater than
                the size of all the NetworkObjectPopulation then all ids are taken and shuffled.
            limit (int): If specified, return the first ``limit`` number of
                IDs from the match result. If limit is greater than the size of all the population
                then all IDs are returned.

        Returns:
            CircuitNodeIds/CircuitEdgeIds: containing the IDs and the populations.
        """
        if not self.population_names:
            raise BluepySnapError("Cannot create CircuitIds for empty population.")

        str_type = f"<U{max(len(pop) for pop in self.population_names)}"
        ids = []
        populations = []
        for pop in self.values():
            pop_ids, name_ids = fun_to_apply(pop)
            pops = np.full_like(pop_ids, fill_value=name_ids, dtype=str_type)
            ids.append(pop_ids)
            populations.append(pops)
        ids = utils.ensure_ids(np.concatenate(ids))
        populations = np.concatenate(populations).astype(str_type)
        res = returned_ids_cls.from_arrays(populations, ids)
        if sample:
            res.sample(sample, inplace=True)
        if limit:
            res.limit(limit, inplace=True)
        return res

    @abc.abstractmethod
    def ids(self, group=None, sample=None, limit=None):
        """Resolves the ids of the NetworkObject."""

    @abc.abstractmethod
    def get(self, group=None, properties=None):
        """Yields the properties of the NetworkObject."""
        ids = self.ids(group)
        properties = utils.ensure_list(properties)
        # We don t convert to set properties itself to keep the column order.
        properties_set = set(properties)

        unknown_props = properties_set - self.property_names
        if unknown_props:
            raise BluepySnapError(f"Unknown properties required: {unknown_props}")

        for name, pop in sorted(self.items()):
            # since ids is sorted, global_pop_ids should be sorted as well
            global_pop_ids = ids.filter_population(name)
            pop_ids = global_pop_ids.get_ids()
            if len(pop_ids) > 0:
                pop_properties = properties_set & pop.property_names
                # Since the columns are passed as Series, index cannot be specified directly.
                # However, it's a bit more performant than converting the Series to numpy arrays.
                pop_df = pd.DataFrame({prop: pop.get(pop_ids, prop) for prop in pop_properties})
                pop_df.index = global_pop_ids.index

                # Sort the columns in the given order
                yield name, pop_df[[p for p in properties if p in pop_properties]]

    @abc.abstractmethod
    def __getstate__(self):
        """Make pickle-able, without storing state of caches."""

    @abc.abstractmethod
    def __setstate__(self, state):
        """Load from pickle state."""



from unittest.mock import patch

import libsonata
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from libsonata import SpikeReader

import bluepysnap.spike_report as test_module
from bluepysnap.bbp import Cell
from bluepysnap.circuit_ids import CircuitNodeIds
from bluepysnap.circuit_ids_types import IDS_DTYPE, CircuitNodeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.simulation import Simulation

from utils import TEST_DATA_DIR, copy_test_data, edit_config


def _create_series(node_ids, index, name="ids"):
    def _get_index(ids):
        return pd.Index(ids, name="times")

    if len(node_ids) == 0:
        # removing warning
        return pd.Series(node_ids, index=_get_index(index), name=name, dtype=np.int64)
    return pd.Series(node_ids, index=_get_index(index), name=name)


class TestSpikeReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.SpikeReport(self.simulation)

    def test_config(self):
        assert isinstance(self.test_obj.config, libsonata._libsonata.Output)
        assert self.test_obj.config.output_dir == str(TEST_DATA_DIR / "reporting")
        assert self.test_obj.config.log_file == "log_spikes.log"
        assert self.test_obj.config.spikes_file == "spikes.h5"
        assert self.test_obj.config.spikes_sort_order.name == "by_time"

    def test_time_start(self):
        assert self.test_obj.time_start == 0.0

    def test_time_stop(self):
        assert self.test_obj.time_stop == 1000.0

    def test_dt(self):
        assert self.test_obj.dt == 0.01

    def test_time_units(self):
        with pytest.raises(NotImplementedError):
            self.test_obj.time_units

    def test_simulation(self):
        assert isinstance(self.test_obj.simulation, Simulation)

    def test_log(self):
        lines = ["Log for spikes.", "Second line."]
        with self.test_obj.log() as log:
            for l, line in enumerate(log):
                assert lines[l] in line

    def test_log2(self):
        with copy_test_data(config="simulation_config.json") as (_, config_path):
            with edit_config(config_path) as config:
                config["output"]["log_file"] = "unknown"

            simulation = Simulation(config_path)
            test_obj = test_module.SpikeReport(simulation)
            with pytest.raises(BluepySnapError):
                with test_obj.log() as log:
                    log.readlines()

    def test___spike_reader(self):
        assert isinstance(self.test_obj._spike_reader, SpikeReader)

    def test_population_names(self):
        assert sorted(self.test_obj.population_names) == ["default", "default2"]

    def test_get_population(self):
        assert isinstance(self.test_obj["default"], test_module.PopulationSpikeReport)

    def test_iter(self):
        assert sorted(list(self.test_obj)) == ["default", "default2"]
        for spikes in self.test_obj:
            isinstance(spikes, test_module.PopulationSpikeReport)

    def test_filter(self):
        filtered = self.test_obj.filter(group=None, t_start=0.3, t_stop=0.7)
        assert filtered.spike_report == self.test_obj
        assert filtered.t_start == 0.3
        assert filtered.t_stop == 0.7
        assert filtered.group is None
        assert isinstance(filtered, test_module.FilteredSpikeReport)
        npt.assert_allclose(filtered.report.index, np.array([0.3, 0.3, 0.7, 0.7]))
        assert filtered.report.columns.tolist() == ["ids", "population"]

        filtered = self.test_obj.filter(group={"other1": ["B"]}, t_start=0.3, t_stop=0.7)
        npt.assert_allclose(filtered.report.index, np.array([0.3]))
        assert filtered.report["population"].unique() == ["default2"]

        filtered = self.test_obj.filter(group={"population": "default2"})
        assert filtered.report["population"].unique() == ["default2"]
        npt.assert_array_equal(sorted(filtered.report["ids"].unique()), [0, 1, 2])

        filtered = self.test_obj.filter(group={"population": "default3"}, t_start=0.3, t_stop=0.6)
        assert len(filtered.report) == 0
        assert set(filtered.report.columns) == {"ids", "population"}
        assert filtered.report.index.name == "times"

        ids = CircuitNodeIds.from_arrays(["default", "default", "default2"], [0, 1, 2])
        filtered = self.test_obj.filter(group=ids)
        npt.assert_array_equal(sorted(filtered.report["ids"].unique()), [0, 1, 2])
        npt.assert_array_equal(
            filtered.report[filtered.report["population"] == "default"]["ids"].unique(), [0, 1]
        )

        filtered = self.test_obj.filter(group=0)
        npt.assert_array_equal(sorted(filtered.report["ids"].unique()), [0])
        npt.assert_array_equal(
            sorted(filtered.report["population"].unique()), ["default", "default2"]
        )

        filtered = self.test_obj.filter(group=[0, 1])
        npt.assert_array_equal(sorted(filtered.report["ids"].unique()), [0, 1])
        npt.assert_array_equal(
            sorted(filtered.report["population"].unique()), ["default", "default2"]
        )


class TestPopulationSpikeReport:
    def setup_method(self):
        self.simulation = Simulation(str(TEST_DATA_DIR / "simulation_config.json"))
        self.test_obj = test_module.SpikeReport(self.simulation)["default"]

    def test__sorted_by(self):
        assert self.test_obj._sorted_by == "by_time"

    def test_name(self):
        assert self.test_obj.name == "default"

    def test_nodes(self):
        node_ids = self.test_obj.get([2], t_start=0.5).to_numpy()[0]
        assert self.test_obj.nodes.get(group=node_ids, properties=Cell.MTYPE) == "L6_Y"
        npt.assert_equal(node_ids.dtype, IDS_DTYPE)

    def test_nodes_invalid_population(self):
        test_obj = test_module.SpikeReport(self.simulation)["default2"]
        test_obj._population_name = "unknown"
        with pytest.raises(BluepySnapError):
            test_obj.nodes

    def test__resolve_nodes(self):
        npt.assert_array_equal(self.test_obj._resolve_nodes({Cell.MTYPE: "L6_Y"}), [1, 2])
        assert self.test_obj._resolve_nodes({Cell.MTYPE: "L2_X"}) == [0]
        npt.assert_array_equal(self.test_obj._resolve_nodes("Node12_L6_Y"), [1, 2])

    def test_get(self):
        tested = self.test_obj.get()
        pdt.assert_series_equal(tested, _create_series([2, 0, 1, 2, 0], [0.1, 0.2, 0.3, 0.7, 1.3]))
        npt.assert_equal(tested.dtype, IDS_DTYPE)

        pdt.assert_series_equal(self.test_obj.get([]), _create_series([], []))
        pdt.assert_series_equal(self.test_obj.get(np.array([])), _create_series([], []))
        pdt.assert_series_equal(self.test_obj.get(()), _create_series([], []))
        pdt.assert_series_equal(self.test_obj.get(2), _create_series([2, 2], [0.1, 0.7]))

        pdt.assert_series_equal(
            self.test_obj.get(CircuitNodeId("default", 2)), _create_series([2, 2], [0.1, 0.7])
        )
        # not in population
        pdt.assert_series_equal(
            self.test_obj.get(CircuitNodeId("default2", 2)), _create_series([], [])
        )

        pdt.assert_series_equal(self.test_obj.get(0, t_start=1.0), _create_series([0], [1.3]))
        pdt.assert_series_equal(self.test_obj.get(0, t_stop=1.0), _create_series([0], [0.2]))
        pdt.assert_series_equal(
            self.test_obj.get(0, t_start=1.0, t_stop=12), _create_series([0], [1.3])
        )
        pdt.assert_series_equal(
            self.test_obj.get(0, t_start=0.1, t_stop=12), _create_series([0, 0], [0.2, 1.3])
        )

        pdt.assert_series_equal(
            self.test_obj.get([2, 0]), _create_series([2, 0, 2, 0], [0.1, 0.2, 0.7, 1.3])
        )

        pdt.assert_series_equal(
            self.test_obj.get([0, 2]), _create_series([2, 0, 2, 0], [0.1, 0.2, 0.7, 1.3])
        )

        pdt.assert_series_equal(
            self.test_obj.get(np.asarray([0, 2])),
            _create_series([2, 0, 2, 0], [0.1, 0.2, 0.7, 1.3]),
        )

        pdt.assert_series_equal(self.test_obj.get([2], t_stop=0.5), _create_series([2], [0.1]))

        pdt.assert_series_equal(self.test_obj.get([2], t_start=0.5), _create_series([2], [0.7]))

        pdt.assert_series_equal(
            self.test_obj.get([2], t_start=0.5, t_stop=0.8), _create_series([2], [0.7])
        )

        pdt.assert_series_equal(
            self.test_obj.get([2, 1], t_start=0.5, t_stop=0.8), _create_series([2], [0.7])
        )

        pdt.assert_series_equal(
            self.test_obj.get([2, 1], t_start=0.2, t_stop=0.8), _create_series([1, 2], [0.3, 0.7])
        )

        pdt.assert_series_equal(
            self.test_obj.get(group={Cell.MTYPE: "L6_Y"}, t_start=0.2, t_stop=0.8),
            _create_series([1, 2], [0.3, 0.7]),
        )

        pdt.assert_series_equal(
            self.test_obj.get(group={Cell.MTYPE: "L2_X"}), _create_series([0, 0], [0.2, 1.3])
        )

        pdt.assert_series_equal(
            self.test_obj.get(group="Layer23"), _create_series([0, 0], [0.2, 1.3])
        )

        # no 0.1, 0.7 from  ("default2", 2)
        ids = CircuitNodeIds.from_arrays(["default", "default", "default2"], [0, 1, 2])
        npt.assert_array_equal(self.test_obj.get(ids), _create_series([0, 1, 0], [0.2, 0.3, 1.3]))

        with pytest.raises(BluepySnapError):
            self.test_obj.get([-1], t_start=0.2)

        with pytest.raises(BluepySnapError):
            self.test_obj.get([0, 2], t_start=-1)

        with pytest.raises(BluepySnapError):
            self.test_obj.get([0, 2], t_start=12)

        with pytest.raises(BluepySnapError):
            self.test_obj.get(4)

    def test_get2(self):
        test_obj = test_module.SpikeReport(self.simulation)["default2"]
        assert test_obj._sorted_by == "by_id"
        pdt.assert_series_equal(
            test_obj.get([2, 0]), _create_series([2, 0, 2, 0], [0.1, 0.2, 0.7, 1.3], name="ids")
        )

        pdt.assert_series_equal(
            test_obj.get([0, 2]), _create_series([2, 0, 2, 0], [0.1, 0.2, 0.7, 1.3], name="ids")
        )

    @patch(
        test_module.__name__ + ".PopulationSpikeReport._resolve_nodes", return_value=np.asarray([4])
    )
    def test_get_not_in_report(self, mock):
        pdt.assert_series_equal(self.test_obj.get(4), _create_series([], []))

    @patch(
        test_module.__name__ + ".PopulationSpikeReport._resolve_nodes",
        return_value=np.asarray([0, 4]),
    )
    def test_get_not_in_report(self, mock):
        pdt.assert_series_equal(self.test_obj.get([0, 4]), _create_series([0, 0], [0.2, 1.3]))

    def test_node_ids(self):
        npt.assert_array_equal(self.test_obj.node_ids, np.array(sorted([0, 1, 2]), dtype=np.int64))



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Edge population access."""
import inspect
from collections.abc import Mapping

import libsonata
import numpy as np
import pandas as pd
from cached_property import cached_property
from more_itertools import first

from bluepysnap import query, utils
from bluepysnap.circuit_ids import CircuitEdgeIds
from bluepysnap.circuit_ids_types import IDS_DTYPE, CircuitEdgeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import DYNAMICS_PREFIX, ConstContainer, Edge


def _is_empty(xs):
    return (xs is not None) and (len(xs) == 0)


def _estimate_range_size(func, node_ids, n=3):
    """Median size of index second level for some node IDs from the provided list."""
    assert len(node_ids) > 0
    if len(node_ids) > n:
        node_ids = np.random.choice(node_ids, size=n, replace=False)
    return np.median([len(func(node_id).ranges) for node_id in node_ids])


class EdgePopulation:
    """Edge population access."""

    def __init__(self, circuit, population_name):
        """Initializes a EdgePopulation object from a EdgeStorage and a population name.

        Args:
            circuit (bluepysnap.Circuit): the circuit object containing the edge population
            population_name (str): the name of the edge population

        Returns:
            EdgePopulation: An EdgePopulation object.
        """
        self._circuit = circuit
        self.name = population_name

    @cached_property
    def _properties(self):
        """Edge population properties."""
        return self._circuit.to_libsonata.edge_population_properties(self.name)

    @property
    def _population(self):
        """Libsonata edge population.

        Not cached because it would keep the hdf5 file open.
        """
        return self._circuit.to_libsonata.edge_population(self.name)

    @staticmethod
    def _resolve_node_ids(nodes, group):
        """Node IDs corresponding to node group filter."""
        if group is None:
            return None
        return nodes.ids(group)

    @property
    def size(self):
        """Population size."""
        return self._population.size

    def _nodes(self, population_name):
        """Returns the NodePopulation corresponding to population."""
        result = self._circuit.nodes[population_name]
        return result

    @property
    def type(self):
        """Population type."""
        return self._properties.type

    @cached_property
    def source(self):
        """Source NodePopulation."""
        return self._nodes(self._population.source)

    @cached_property
    def target(self):
        """Target NodePopulation."""
        return self._nodes(self._population.target)

    @cached_property
    def _attribute_names(self):
        return set(self._population.attribute_names)

    @cached_property
    def _dynamics_params_names(self):
        return set(utils.add_dynamic_prefix(self._population.dynamics_attribute_names))

    @property
    def _topology_property_names(self):
        return {Edge.SOURCE_NODE_ID, Edge.TARGET_NODE_ID}

    @property
    def config(self):
        """Access the configuration for the population.

        This configuration is extended with:

        * 'components' of the circuit config
        * 'edges_file': the path the h5 file containing the population.
        """
        return self._circuit.get_edge_population_config(self.name)

    @property
    def property_names(self):
        """Set of available edge properties.

        Notes:
            Properties are a combination of the group attributes, the dynamics_params and the
            topology properties.
        """
        return self._attribute_names | self._dynamics_params_names | self._topology_property_names

    @cached_property
    def property_dtypes(self):
        """Returns the dtypes of all the properties.

        Returns:
            pandas.Series: series indexed by field name with the corresponding dtype as value.
        """
        return self.get([0], list(self.property_names)).dtypes.sort_index()

    def container_property_names(self, container):
        """Lists the ConstContainer properties shared with the EdgePopulation.

        Args:
            container (ConstContainer): a container class for edge properties.

        Returns:
            list: A list of strings corresponding to the properties that you can use from the
            container class

        Examples:
            >>> from bluepysnap.sonata_constants import Edge
            >>> print(my_edge_population.container_property_names(Edge))
            >>> ["AXONAL_DELAY", "SYN_WEIGHT"] # values you can use with my_edge_population
        """
        if not inspect.isclass(container) or not issubclass(container, ConstContainer):
            raise BluepySnapError("'container' must be a subclass of ConstContainer")
        in_file = self.property_names
        return [k for k in container.key_set() if container.get(k) in in_file]

    def _get_property(self, prop, selection):
        if prop == Edge.SOURCE_NODE_ID:
            result = utils.ensure_ids(self._population.source_nodes(selection))
        elif prop == Edge.TARGET_NODE_ID:
            result = utils.ensure_ids(self._population.target_nodes(selection))
        elif prop in self._attribute_names:
            result = self._population.get_attribute(prop, selection)
        elif prop in self._dynamics_params_names:
            result = self._population.get_dynamics_attribute(
                prop.split(DYNAMICS_PREFIX)[1], selection
            )
        else:
            raise BluepySnapError(f"No such property: {prop}")
        return result

    def _get(self, selection, properties=None):
        """Get an array of edge IDs or DataFrame with edge properties."""
        edge_ids = utils.ensure_ids(selection.flatten())
        if properties is None:
            return edge_ids

        if utils.is_iterable(properties):
            if len(edge_ids) == 0:
                result = pd.DataFrame(columns=properties)
            else:
                result = pd.DataFrame(index=edge_ids)
                for p in properties:
                    result[p] = self._get_property(p, selection)
        else:
            if len(edge_ids) == 0:
                result = pd.Series(name=properties, dtype=np.float64)
            else:
                result = pd.Series(
                    self._get_property(properties, selection), index=edge_ids, name=properties
                )

        return result

    def _edge_ids_by_filter(self, queries, raise_missing_prop):
        """Return edge IDs if their properties match the `queries` dict.

        `props` values could be:
            pairs (range match for floating dtype fields)
            scalar or iterables (exact or "one of" match for other fields)

        You can use the special operators '$or' and '$and' also to combine different queries
        together.

        Examples:
            >>> self._edge_ids_by_filter({ Edge.POST_SECTION_ID: (0, 1),
            >>>                            Edge.AXONAL_DELAY: (.5, 2.) })
            >>> self._edge_ids_by_filter({'$or': [{ Edge.PRE_X_CENTER: [2, 3]},
            >>>                              { Edge.POST_SECTION_POS: (0, 1),
            >>>                              Edge.SYN_WEIGHT: (0.,1.4) }]})

        """
        properties = query.get_properties(queries)
        unknown_props = properties - self.property_names
        if raise_missing_prop and unknown_props:
            raise BluepySnapError(f"Unknown edge properties: {unknown_props}")
        res = []
        ids = self.ids(None)
        chunk_size = int(1e8)
        for chunk in np.array_split(ids, 1 + len(ids) // chunk_size):
            data = self.get(chunk, properties - unknown_props)
            res.extend(chunk[query.resolve_ids(data, self.name, queries)])
        return np.array(res, dtype=IDS_DTYPE)

    def ids(self, group=None, limit=None, sample=None, raise_missing_property=True):
        """Edge IDs corresponding to edges ``edge_ids``.

        Args:
            group (None/int/CircuitEdgeId/CircuitEdgeIds/sequence): Which IDs will be
                returned depends on the type of the ``group`` argument:

                - ``None``: return all IDs.
                - ``int``, ``CircuitEdgeId``: return a single edge ID.
                - ``CircuitEdgeIds`` return IDs of edges the edge population in an array.
                - ``sequence``: return IDs of edges in an array.

            sample (int): If specified, randomly choose ``sample`` number of
                IDs from the match result. If the size of the sample is greater than
                the size of the EdgePopulation then all ids are taken and shuffled.

            limit (int): If specified, return the first ``limit`` number of
                IDs from the match result. If limit is greater than the size of the population
                all node IDs are returned.

            raise_missing_property (bool): if True, raises if a property is not listed in this
                population. Otherwise the ids are just not selected if a property is missing.

        Returns:
            numpy.array: A numpy array of IDs.
        """
        if group is None:
            result = self._population.select_all().flatten()
        elif isinstance(group, CircuitEdgeIds):
            result = group.filter_population(self.name).get_ids()
        elif isinstance(group, np.ndarray):
            result = group
        elif isinstance(group, Mapping):
            result = self._edge_ids_by_filter(
                queries=group, raise_missing_prop=raise_missing_property
            )
        else:
            result = utils.ensure_list(group)
            # test if first value is a CircuitEdgeId if yes then all values must be CircuitEdgeId
            if isinstance(first(result, None), CircuitEdgeId):
                try:
                    result = [cid.id for cid in result if cid.population == self.name]
                except AttributeError as e:
                    raise BluepySnapError(
                        "All values from a list must be of type int or CircuitEdgeId."
                    ) from e
        if sample is not None:
            if len(result) > 0:
                result = np.random.choice(result, min(sample, len(result)), replace=False)
        if limit is not None:
            result = result[:limit]
        return utils.ensure_ids(result)

    def get(self, edge_ids, properties):
        """Edge properties as pandas DataFrame.

        Args:
            edge_ids (array-like): array-like of edge IDs
            properties (str/list): an edge property name or a list of edge property names

        Returns:
            pandas.Series/pandas.DataFrame:
                - A pandas Series indexed by edge IDs if ``properties`` is scalar.
                - A pandas DataFrame indexed by edge IDs if ``properties`` is list.

        Notes:
            The EdgePopulation.property_names function will give you all the usable properties
            for the `properties` argument.
        """
        edge_ids = self.ids(edge_ids)
        selection = libsonata.Selection(edge_ids)
        return self._get(selection, properties)

    def positions(self, edge_ids, side, kind):
        """Edge positions as a pandas DataFrame.

        Args:
            edge_ids (array-like): array-like of edge IDs
            side (str): ``afferent`` or ``efferent``
            kind (str): ``center`` or ``surface``

        Returns:
            Pandas Dataframe with ('x', 'y', 'z') columns indexed by edge IDs.
        """
        assert side in ("afferent", "efferent")
        assert kind in ("center", "surface")
        props = {f"{side}_{kind}_{p}": p for p in ["x", "y", "z"]}
        result = self.get(edge_ids, list(props))
        result.rename(columns=props, inplace=True)
        result.sort_index(axis=1, inplace=True)
        return result

    def afferent_nodes(self, target, unique=True):
        """Get afferent node IDs for given target ``node_id``.

        Notes:
            Afferent nodes are nodes projecting an outgoing edge to one of the ``target`` node.

        Args:
            target (CircuitNodeIds/int/sequence/str/mapping/None): the target you want to resolve
            and use as target nodes.
            unique (bool): If ``True``, return only unique afferent node IDs.

        Returns:
            numpy.ndarray: Afferent node IDs for all the targets.
        """
        if target is not None:
            selection = self._population.afferent_edges(self._resolve_node_ids(self.target, target))
        else:
            selection = self._population.select_all()
        result = self._population.source_nodes(selection)
        if unique:
            result = np.unique(result)
        return utils.ensure_ids(result)

    def efferent_nodes(self, source, unique=True):
        """Get efferent node IDs for given source ``node_id``.

        Notes:
            Efferent nodes are nodes receiving an incoming edge from one of the ``source`` node.

        Args:
            source (CircuitNodeIds/int/sequence/str/mapping/None): the source you want to resolve
                and use as source nodes.
            unique (bool): If ``True``, return only unique afferent node IDs.

        Returns:
            numpy.ndarray: Efferent node IDs for all the sources.
        """
        if source is not None:
            selection = self._population.efferent_edges(self._resolve_node_ids(self.source, source))
        else:
            selection = self._population.select_all()
        result = self._population.target_nodes(selection)
        if unique:
            result = np.unique(result)
        return utils.ensure_ids(result)

    def pathway_edges(self, source=None, target=None, properties=None):
        """Get edges corresponding to ``source`` -> ``target`` connections.

        Args:
            source (CircuitNodeIds/int/sequence/str/mapping/None): source node group
            target (CircuitNodeIds/int/sequence/str/mapping/None): target node group
            properties: None / edge property name / list of edge property names

        Returns:
            - List of edge IDs, if ``properties`` is None;
            - Pandas Series indexed by edge IDs if ``properties`` is string;
            - Pandas DataFrame indexed by edge IDs if ``properties`` is list.
        """
        if source is None and target is None:
            raise BluepySnapError("Either `source` or `target` should be specified")

        source_node_ids = self._resolve_node_ids(self.source, source)
        target_edge_ids = self._resolve_node_ids(self.target, target)

        if source_node_ids is None:
            selection = self._population.afferent_edges(target_edge_ids)
        elif target_edge_ids is None:
            selection = self._population.efferent_edges(source_node_ids)
        else:
            selection = self._population.connecting_edges(source_node_ids, target_edge_ids)

        if properties:
            return self._get(selection, properties)
        return utils.ensure_ids(selection.flatten())

    def afferent_edges(self, node_id, properties=None):
        """Get afferent edges for given ``node_id``.

        Args:
            node_id (CircuitNodeIds/int/sequence/str/mapping/None) : Target node ID.
            properties: An edge property name, a list of edge property names, or None.

        Returns:
            pandas.Series/pandas.DataFrame/list:
                - A pandas Series indexed by edge ID if ``properties`` is a string.
                - A pandas DataFrame indexed by edge ID if ``properties`` is a list.
                - A list of edge IDs, if ``properties`` is None.
        """
        return self.pathway_edges(source=None, target=node_id, properties=properties)

    def efferent_edges(self, node_id, properties=None):
        """Get efferent edges for given ``node_id``.

        Args:
            node_id (CircuitNodeIds/int/sequence/str/mapping/None): source node ID
            properties: None / edge property name / list of edge property names

        Returns:
            - List of edge IDs, if ``properties`` is None;
            - Pandas Series indexed by edge IDs if ``properties`` is string;
            - Pandas DataFrame indexed by edge IDs if ``properties`` is list.
        """
        return self.pathway_edges(source=node_id, target=None, properties=properties)

    def pair_edges(self, source_node_id, target_node_id, properties=None):
        """Get edges corresponding to ``source_node_id`` -> ``target_node_id`` connection.

        Args:
            source_node_id (CircuitNodeIds/int/sequence/str/mapping/None): source node ID
            target_node_id (CircuitNodeIds/int/sequence/str/mapping/None): target node ID
            properties: None / edge property name / list of edge property names

        Returns:
            - List of edge IDs, if ``properties`` is None;
            - Pandas Series indexed by edge IDs if ``properties`` is string;
            - Pandas DataFrame indexed by edge IDs if ``properties`` is list.
        """
        return self.pathway_edges(
            source=source_node_id, target=target_node_id, properties=properties
        )

    def _iter_connections(self, source_node_ids, target_node_ids, unique_node_ids, shuffle):
        """Iterate through `source_node_ids` -> `target_node_ids` connections."""

        # pylint: disable=too-many-branches,too-many-locals
        def _optimal_direction():
            """Choose between source and target node IDs for iterating."""
            if target_node_ids is None and source_node_ids is None:
                raise BluepySnapError("Either `source` or `target` should be specified")
            if source_node_ids is None:
                return "target"
            if target_node_ids is None:
                return "source"
            else:
                # Checking the indexing 'direction'. One direction has contiguous indices.
                range_size_source = _estimate_range_size(
                    self._population.efferent_edges, source_node_ids
                )
                range_size_target = _estimate_range_size(
                    self._population.afferent_edges, target_node_ids
                )
                return "source" if (range_size_source < range_size_target) else "target"

        if _is_empty(source_node_ids) or _is_empty(target_node_ids):
            return

        direction = _optimal_direction()
        if direction == "target":
            primary_node_ids, secondary_node_ids = target_node_ids, source_node_ids
            get_connected_node_ids = self.afferent_nodes
        else:
            primary_node_ids, secondary_node_ids = source_node_ids, target_node_ids
            get_connected_node_ids = self.efferent_nodes

        primary_node_ids = np.unique(primary_node_ids)
        if shuffle:
            np.random.shuffle(primary_node_ids)

        if secondary_node_ids is not None:
            secondary_node_ids = np.unique(secondary_node_ids)

        secondary_node_ids_used = set()

        for key_node_id in primary_node_ids:
            connected_node_ids = get_connected_node_ids(key_node_id, unique=False)
            # [[secondary_node_id, count], ...]
            connected_node_ids_with_count = np.stack(
                np.unique(connected_node_ids, return_counts=True)
            ).transpose()
            # np.stack(uint64, int64) -> float64
            connected_node_ids_with_count = connected_node_ids_with_count.astype(np.uint32)
            if secondary_node_ids is not None:
                mask = np.in1d(
                    connected_node_ids_with_count[:, 0], secondary_node_ids, assume_unique=True
                )
                connected_node_ids_with_count = connected_node_ids_with_count[mask]
            if shuffle:
                np.random.shuffle(connected_node_ids_with_count)

            for conn_node_id, edge_count in connected_node_ids_with_count:
                if unique_node_ids and (conn_node_id in secondary_node_ids_used):
                    continue
                if direction == "target":
                    yield conn_node_id, key_node_id, edge_count
                else:
                    yield key_node_id, conn_node_id, edge_count
                if unique_node_ids:
                    secondary_node_ids_used.add(conn_node_id)
                    break

    def iter_connections(
        self,
        source=None,
        target=None,
        unique_node_ids=False,
        shuffle=False,
        return_edge_ids=False,
        return_edge_count=False,
    ):
        """Iterate through ``source`` -> ``target`` connections.

        Args:
            source (CircuitNodeIds/int/sequence/str/mapping/None): source node group
            target (CircuitNodeIds/int/sequence/str/mapping/None): target node group
            unique_node_ids: if True, no node ID will be used more than once as source or
                target for edges. Careful, this flag does not provide unique (source, target)
                pairs but unique node IDs.
            shuffle: if True, result order would be (somewhat) randomized
            return_edge_count: if True, edge count is added to yield result
            return_edge_ids: if True, edge ID list is added to yield result

        ``return_edge_count`` and ``return_edge_ids`` are mutually exclusive.

        Yields:
            - (source_node_id, target_node_id, edge_ids) if ``return_edge_ids`` is True;
            - (source_node_id, target_node_id, edge_count) if ``return_edge_count`` is True;
            - (source_node_id, target_node_id) otherwise.
        """
        if return_edge_ids and return_edge_count:
            raise BluepySnapError(
                "`return_edge_count` and `return_edge_ids` are mutually exclusive"
            )

        source_node_ids = self._resolve_node_ids(self.source, source)
        target_node_ids = self._resolve_node_ids(self.target, target)

        it = self._iter_connections(source_node_ids, target_node_ids, unique_node_ids, shuffle)

        if return_edge_count:
            return it
        elif return_edge_ids:
            add_edge_ids = lambda x: (x[0], x[1], self.pair_edges(x[0], x[1]))
            return map(add_edge_ids, it)
        else:
            omit_edge_count = lambda x: x[:2]
            return map(omit_edge_count, it)

    @property
    def h5_filepath(self):
        """Get the H5 edges file associated with population."""
        return self.config["edges_file"]

    @cached_property
    def spatial_synapse_index(self):
        """Access to edges spatial index."""
        try:
            from spatial_index import open_index
        except ImportError as e:
            raise BluepySnapError(
                (
                    "Spatial index is for now only available internally to BBP. "
                    "It requires `spatial_index`, an internal package."
                )
            ) from e

        index_dir = self._properties.spatial_synapse_index_dir
        if not index_dir:
            raise BluepySnapError(f"It appears {self.name} does not have synapse indices")
        return open_index(index_dir)

    def __getstate__(self):
        """Make EdgePopulation pickle-able, without storing state of caches."""
        return self._circuit, self.name

    def __setstate__(self, state):
        """Load from pickle state."""
        circuit, population_name = state
        self.__init__(circuit, population_name)



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Spike report access."""

from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pandas as pd
from cached_property import cached_property
from libsonata import SonataError, SpikeReader

import bluepysnap._plotting
from bluepysnap.circuit_ids_types import IDS_DTYPE
from bluepysnap.exceptions import BluepySnapError


def _get_reader(config):
    path = str(Path(config.output_dir) / config.spikes_file)
    return SpikeReader(path)


def _collect_spikes(spike_report):
    return {
        population: PopulationSpikeReport(spike_report, population)
        for population in spike_report.population_names
    }


class PopulationSpikeReport:
    """Access to PopulationSpikeReport data."""

    def __init__(self, spike_report, population_name):
        """Initializes a PopulationSpikeReport object from a SpikeReport.

        Args:
            spike_report (SpikeReport): SpikeReport containing this spike report population.
            population_name (str): the population name corresponding to this report.

        Returns:
            PopulationSpikeReport: A PopulationSpikeReport object.
        """
        self.spike_report = spike_report
        self._spike_population = _get_reader(self.spike_report.config)[population_name]
        self._population_name = population_name

    @property
    def _sorted_by(self):
        """Access to the sorting attribute.

        Returns:
            str: the type of sorting used for this spike report.

            Returned values are:

               - ``'by_id'`` if the report is sorted by population node_ids
               - ``'by_time'`` if the report is sorted by timestamps
               - ``'none'`` if not sorted.
        """
        return self._spike_population.sorting

    @property
    def name(self):
        """Return the name of the population."""
        return self._population_name

    @cached_property
    def nodes(self):
        """Return the NodePopulation corresponding to this spike report."""
        return self.spike_report.simulation.circuit.nodes[self._population_name]

    def _resolve_nodes(self, group):
        """Transform a node group into a node_id array."""
        return self.nodes.ids(group=group)

    def get(self, group=None, t_start=None, t_stop=None):
        """Fetch spikes from the report.

        Args:
            group (None/int/list/np.array/dict): Get spikes filtered by group. See NodePopulation.
            t_start (float): Include only spikes occurring at or after this time.
            t_stop (float): Include only spikes occurring at or before this time.

        Returns:
            pandas.Series: return spiking node_ids indexed by sorted spike time.
        """
        node_ids = self._resolve_nodes(group).tolist()

        series_name = "ids"
        try:
            res = self._spike_population.get(node_ids=node_ids, tstart=t_start, tstop=t_stop)
        except SonataError as e:
            raise BluepySnapError(e) from e

        if not res:
            return pd.Series(
                data=[], index=pd.Index([], name="times"), name=series_name, dtype=IDS_DTYPE
            )

        # pylint: disable=unsubscriptable-object
        res = pd.DataFrame(data=res, columns=[series_name, "times"]).set_index("times")[series_name]
        if self._sorted_by != "by_time":
            res.sort_index(inplace=True)
        return res.astype(IDS_DTYPE)

    @cached_property
    def node_ids(self):
        """Returns the node ids present in the report.

        Returns:
            np.Array: Numpy array containing the node_ids included in the report
        """
        return np.unique(self.get())


class FilteredSpikeReport:
    """Access to filtered SpikeReport data."""

    def __init__(self, spike_report, group=None, t_start=None, t_stop=None):
        """Initialize a FilteredSpikeReport.

        A FilteredSpikeReport is a lazy and cached object which contains the filtered data
        from all the populations of a report.

        Args:
            spike_report (SpikeReport): The SpikeReport to filter.
            group (None/int/list/np.array/dict): Get spikes filtered by group. See NodePopulation.
            t_start (float): Include only frames occurring at or after this time.
            t_stop (float): Include only frames occurring at or before this time.

        Returns:
            FilteredSpikeReport: A FilteredSpikeReport object.
        """
        self.spike_report = spike_report
        self.group = group
        self.t_start = t_start
        self.t_stop = t_stop

    @cached_property
    def report(self):
        """Access to the report data.

        Returns:
            pandas.DataFrame: A DataFrame containing the data from the report. Row's indices are the
            different timestamps and the columns are ids and population names.
        """
        dfs = []
        for population in self.spike_report.population_names:
            spikes = self.spike_report[population]
            ids = spikes.nodes.ids(group=self.group, raise_missing_property=False)
            data = spikes.get(group=ids, t_start=self.t_start, t_stop=self.t_stop).to_frame()
            data["population"] = np.full(len(data), population)

            dfs.append(data)

        if all(df.empty for df in dfs):
            res = dfs[0][:0]
        else:
            res = pd.concat([df for df in dfs if not df.empty])

        res["population"] = res["population"].astype("category")

        return res.sort_index()

    # pylint: disable=protected-access
    raster = bluepysnap._plotting.spike_raster
    firing_rate_histogram = bluepysnap._plotting.spikes_firing_rate_histogram
    isi = bluepysnap._plotting.spikes_isi
    firing_animation = bluepysnap._plotting.spikes_firing_animation


class SpikeReport:
    """Access to SpikeReport data."""

    def __init__(self, simulation):
        """Initializes a SpikeReport object from a simulation object.

        Args:
            simulation (Simulation): Simulation containing this spike report.

        Returns:
            SpikeReport: A SpikeReport object.
        """
        self._simulation = simulation

    @property
    def config(self):
        """Access to the spike 'output' config part."""
        return self._simulation.output

    @property
    def time_start(self):
        """Returns the starting time of the simulation. Default is zero."""
        return self._simulation.time_start

    @property
    def time_stop(self):
        """Returns the stopping time of the simulation."""
        return self._simulation.time_stop

    @property
    def dt(self):
        """Returns the frequency of reporting in milliseconds."""
        return self._simulation.dt

    @property
    def time_units(self):
        """Returns the time unit of reporting."""
        raise NotImplementedError

    @property
    def simulation(self):
        """Return the Simulation object related to this spike report."""
        return self._simulation

    @contextmanager
    def log(self):
        """Context manager for the spike log file."""
        path = Path(self.config.output_dir) / self.config.log_file
        if not path.exists():
            raise BluepySnapError("Cannot find the log file for the spike report.")
        with path.open("r", encoding="utf-8") as f:
            yield f

    @cached_property
    def _spike_reader(self):
        """Access to the libsonata SpikeReader."""
        return _get_reader(self.config)

    @cached_property
    def population_names(self):
        """Returns the population names included in this report."""
        return sorted(self._spike_reader.get_population_names())

    @cached_property
    def _population(self):
        """Collect the different PopulationSpikeReports."""
        return _collect_spikes(self)

    def __getitem__(self, population_name):
        """Access the PopulationSpikeReport corresponding to the population 'population_name'."""
        return self._population[population_name]

    def __iter__(self):
        """Allows iteration over the different PopulationSpikeReports."""
        return iter(self._population)

    def filter(self, group=None, t_start=None, t_stop=None):
        """Returns a FilteredSpikeReport.

        A FilteredSpikeReport is a lazy and cached object which contains the filtered data
        from all the populations of a report.

        Args:
            group (None/int/list/np.array/dict): Get spikes filtered by group. See NodePopulation.
            t_start (float): Include only frames occurring at or after this time.
            t_stop (float): Include only frames occurring at or before this time.

        Returns:
            FilteredSpikeReport: A FilteredSpikeReport object.
        """
        return FilteredSpikeReport(self, group, t_start, t_stop)



import bluepysnap.circuit_ids_types as test_module


class TestCircuitNodeId:
    def setup_method(self):
        self.test_obj = test_module.CircuitNodeId("pop", 1)

    def test_init(self):
        assert isinstance(self.test_obj, test_module.CircuitNodeId)
        assert isinstance(self.test_obj, tuple)

    def test_accessors(self):
        assert self.test_obj.population == "pop"
        assert self.test_obj.id == 1


class TestCircuitEdgeId:
    def setup_method(self):
        self.test_obj = test_module.CircuitEdgeId("pop", 1)

    def test_init(self):
        assert isinstance(self.test_obj, test_module.CircuitEdgeId)
        assert isinstance(self.test_obj, tuple)

    def test_accessors(self):
        assert self.test_obj.population == "pop"
        assert self.test_obj.id == 1



# Copyright (c) 2023, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""CircuitIds types."""

from collections import namedtuple

import numpy as np

# dtypes for the different node and edge ids. We are using np.int64 to avoid the infamous
# https://github.com/numpy/numpy/issues/15084 numpy problem. This type needs to be used for
# all returned node or edge ids.
IDS_DTYPE = np.int64

CircuitNodeId = namedtuple("CircuitNodeId", ("population", "id"))
CircuitEdgeId = namedtuple("CircuitEdgeId", ("population", "id"))



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Access to node set data.

For more information see:
https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md#node-sets-file
"""

import copy
import json

import libsonata

from bluepysnap import utils
from bluepysnap.exceptions import BluepySnapError


class NodeSet:
    """Access to single node set."""

    def __init__(self, node_sets, name):
        """Initializes a single node set object.

        Args:
            node_sets (libsonata.NodeSets): libsonata NodeSets instance.
            name (str): name of the node set.

        Returns:
            NodeSet: A NodeSet object.
        """
        self._node_sets = node_sets
        self._name = name

    def get_ids(self, population, raise_missing_property=True):
        """Get the resolved node set as ids."""
        try:
            return self._node_sets.materialize(self._name, population).flatten()
        except libsonata.SonataError as e:
            if not raise_missing_property and "No such attribute" in e.args[0]:
                return []
            raise BluepySnapError(*e.args) from e


class NodeSets:
    """Access to node sets data."""

    def __init__(self, content, instance):
        """Initializes a node set object from a node sets file.

        Args:
            content (dict): Node sets as a dictionary.
            instance (libsonata.NodeSets): ``libsonata`` node sets instance.

        Returns:
            NodeSets: A NodeSets object.
        """
        self._content = content
        self._instance = instance

    @classmethod
    def from_file(cls, filepath):
        """Create NodeSets instance from a file."""
        content = utils.load_json(filepath)
        instance = libsonata.NodeSets.from_file(filepath)
        return cls(content, instance)

    @classmethod
    def from_string(cls, content):
        """Create NodeSets instance from a JSON string."""
        instance = libsonata.NodeSets(content)
        content = json.loads(content)
        return cls(content, instance)

    @classmethod
    def from_dict(cls, content):
        """Create NodeSets instance from a dict."""
        return cls.from_string(json.dumps(content))

    @property
    def content(self):
        """Access (a copy of) the node sets contents."""
        return copy.deepcopy(self._content)

    @property
    def to_libsonata(self):
        """Libsonata instance of the NodeSets."""
        return self._instance

    def update(self, node_sets):
        """Update the contents of the node set.

        Args:
            node_sets (bluepysnap.NodeSets): The node set to extend this node set with.

        Returns:
            set: Names of any overwritten node sets.
        """
        if isinstance(node_sets, NodeSets):
            overwritten = self._instance.update(node_sets.to_libsonata)
            self._content.update(node_sets.content)
            return overwritten

        raise BluepySnapError(
            f"Unexpected type: '{type(node_sets).__name__}' "
            f"(expected: '{self.__class__.__name__}')"
        )

    def __contains__(self, name):
        """Check if node set exists."""
        if isinstance(name, str):
            return name in self._instance.names

        raise BluepySnapError(f"Unexpected type: '{type(name).__name__}' (expected: 'str')")

    def __getitem__(self, name):
        """Return a node set instance for the given node set name."""
        if name not in self:
            raise BluepySnapError(f"Undefined node set: '{name}'")
        return NodeSet(self._instance, name)

    def __iter__(self):
        """Iter through the different node sets names."""
        return iter(self._instance.names)



import sys
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

import bluepysnap._plotting as test_module
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.simulation import Simulation
from bluepysnap.spike_report import FilteredSpikeReport, SpikeReport

from utils import TEST_DATA_DIR

# NOTE: The tests here are primarily to make sure all the code is covered and deprecation warnings,
# etc. are raised. They don't ensure nor really test the correctness of the functionality.


def test__get_pyplot():
    with patch.dict(sys.modules, {"matplotlib.pyplot": None}):
        with pytest.raises(ImportError):
            test_module._get_pyplot()

    import matplotlib.pyplot

    plt_test = test_module._get_pyplot()
    assert plt_test is matplotlib.pyplot


def _get_filtered_spike_report():
    return Simulation(TEST_DATA_DIR / "simulation_config.json").spikes.filter()


def _get_filtered_frame_report():
    return Simulation(TEST_DATA_DIR / "simulation_config.json").reports["soma_report"].filter()


def test_spikes_firing_rate_histogram():
    with pytest.raises(BluepySnapError, match="Invalid time_binsize"):
        test_module.spikes_firing_rate_histogram(filtered_report=None, time_binsize=0)

    filtered_report = _get_filtered_spike_report()
    ax = test_module.spikes_firing_rate_histogram(filtered_report)
    assert ax.xaxis.label.get_text() == "Time [ms]"
    assert ax.yaxis.label.get_text() == "PSTH [Hz]"

    ax.xaxis.label.set_text("Fake X")
    ax.yaxis.label.set_text("Fake Y")

    ax = test_module.spikes_firing_rate_histogram(filtered_report, ax=ax)
    assert ax.xaxis.label.get_text() == "Fake X"
    assert ax.yaxis.label.get_text() == "Fake Y"


def test_spike_raster():
    filtered_report = _get_filtered_spike_report()

    test_module.spike_raster(filtered_report)
    test_module.spike_raster(filtered_report, y_axis="y")

    ax = test_module.spike_raster(filtered_report, y_axis="mtype")

    assert ax.xaxis.label.get_text() == "Time [ms]"
    assert ax.yaxis.label.get_text() == "mtype"

    ax.xaxis.label.set_text("Fake X")
    ax.yaxis.label.set_text("Fake Y")

    ax = test_module.spike_raster(filtered_report, y_axis="mtype", ax=ax)
    assert ax.xaxis.label.get_text() == "Fake X"
    assert ax.yaxis.label.get_text() == "Fake Y"

    # Have error raised in node_population get
    filtered_report.spike_report["default"].nodes.get = Mock(
        side_effect=BluepySnapError("Fake error")
    )
    test_module.spike_raster(filtered_report, y_axis="mtype")


def test_spikes_isi():
    with pytest.raises(BluepySnapError, match="Invalid binsize"):
        test_module.spikes_isi(filtered_report=None, binsize=0)

    filtered_report = _get_filtered_spike_report()

    ax = test_module.spikes_isi(filtered_report)
    assert ax.xaxis.label.get_text() == "Interspike interval [ms]"
    assert ax.yaxis.label.get_text() == "Bin weight"

    ax = test_module.spikes_isi(filtered_report, use_frequency=True, binsize=42)
    assert ax.xaxis.label.get_text() == "Frequency [Hz]"
    assert ax.yaxis.label.get_text() == "Bin weight"

    ax.xaxis.label.set_text("Fake X")
    ax.yaxis.label.set_text("Fake Y")
    ax = test_module.spikes_isi(filtered_report, use_frequency=True, binsize=42, ax=ax)
    assert ax.xaxis.label.get_text() == "Fake X"
    assert ax.yaxis.label.get_text() == "Fake Y"

    with patch.object(test_module.np, "concatenate", Mock(return_value=[])):
        with pytest.raises(BluepySnapError, match="No data to display"):
            test_module.spikes_isi(filtered_report)


def test_spikes_firing_animation(tmp_path):
    with pytest.raises(BluepySnapError, match="Fake is not a valid axis"):
        test_module.spikes_firing_animation(filtered_report=None, x_axis="Fake")

    with pytest.raises(BluepySnapError, match="Fake is not a valid axis"):
        test_module.spikes_firing_animation(filtered_report=None, y_axis="Fake")

    filtered_report = _get_filtered_spike_report()
    anim, ax = test_module.spikes_firing_animation(filtered_report, dt=0.2)
    assert ax.title.get_text() == "time = 0.1ms"

    # convert to video to have `update_animation` called
    anim.save(tmp_path / "test.gif")

    ax.title.set_text("Fake Title")
    anim, ax = test_module.spikes_firing_animation(filtered_report, dt=0.2, ax=ax)
    assert ax.title.get_text() == "Fake Title"
    anim.save(tmp_path / "test.gif")

    # Have error raised in node_population get
    filtered_report.spike_report["default"].nodes.get = Mock(
        side_effect=BluepySnapError("Fake error")
    )

    anim, _ = test_module.spikes_firing_animation(filtered_report, dt=0.2)
    anim.save(tmp_path / "test.gif")


def test_frame_trace():
    with pytest.raises(BluepySnapError, match="Unknown plot_type Fake."):
        test_module.frame_trace(filtered_report=None, plot_type="Fake", ax="also fake")

    filtered_report = _get_filtered_frame_report()
    test_module.frame_trace(filtered_report)
    ax = test_module.frame_trace(filtered_report, plot_type="all")

    assert ax.xaxis.label.get_text() == "Time [ms]"
    assert ax.yaxis.label.get_text() == "Voltage [mV]"

    ax.xaxis.label.set_text("Fake X")
    ax.yaxis.label.set_text("Fake Y")

    ax = test_module.frame_trace(filtered_report, plot_type="all", ax=ax)

    assert ax.xaxis.label.get_text() == "Fake X"
    assert ax.yaxis.label.get_text() == "Fake Y"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

import importlib

# -- Project information -----------------------------------------------------

project = "Blue Brain SNAP"
author = "Blue Brain Project, EPFL"

release = importlib.metadata.distribution("bluepysnap").version
version = release

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# ensure a useful title is used
html_title = "Blue Brain SNAP"

# hide source links
html_show_sourcelink = False

# set the theme settings
html_theme_options = {
    "repo_url": "https://github.com/BlueBrain/snap/",
    "repo_name": "BlueBrain/snap",
}

# autodoc settings
autodoc_default_options = {
    "members": True,
}

autoclass_content = "both"

autodoc_mock_imports = ["libsonata"]

# autosummary settings
autosummary_generate = True

suppress_warnings = [
    "autosectionlabel.*",
]



# Copyright (c) 2020, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Simulation access."""

import warnings
from pathlib import Path

from cached_property import cached_property

from bluepysnap.config import SimulationConfig
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.input import get_simulation_inputs
from bluepysnap.node_sets import NodeSets


def _collect_frame_reports(sim):
    """Collect the different frame reports."""
    res = {}
    for name in sim.to_libsonata.list_report_names:
        report = sim.to_libsonata.report(name)
        report_type = report.sections.name
        if report_type == "soma":
            from bluepysnap.frame_report import SomaReport

            cls = SomaReport
        elif report_type == "all":
            from bluepysnap.frame_report import CompartmentReport

            cls = CompartmentReport
        else:
            raise BluepySnapError(f"Report {name}: format {report_type} not yet supported.")

        res[name] = cls(sim, name)
    return res


def _warn_on_overwritten_node_sets(overwritten, print_max=10):
    """Helper function to warn and print overwritten nodesets."""
    if (n := len(overwritten)) > 0:
        names = ", ".join(list(overwritten)[:print_max]) + (", ..." if n > print_max else "")
        warnings.warn(
            f"Simulation node sets overwrite {n} node set(s) in Circuit node sets: {names}",
            RuntimeWarning,
        )


class Simulation:
    """Access to Simulation data."""

    def __init__(self, config):
        """Initializes a simulation object from a SONATA simulation config file.

        Args:
            config (str): Path to a SONATA simulation config file.

        Returns:
            Simulation: A Simulation object.
        """
        self._simulation_config_path = str(Path(config).absolute())
        self._config = SimulationConfig.from_config(config)

    @property
    def to_libsonata(self):
        """Libsonata instance of the config."""
        return self._config.to_libsonata

    @property
    def config(self):
        """Simulation config dictionary."""
        return self._config.to_dict()

    @cached_property
    def circuit(self):
        """Access to the circuit used for the simulation."""
        from bluepysnap.circuit import Circuit

        if not Path(self.to_libsonata.network).is_file():
            raise BluepySnapError(f"'network' file not found: {self.to_libsonata.network}")
        return Circuit(self.to_libsonata.network)

    @property
    def output(self):
        """Access the output section."""
        return self.to_libsonata.output

    @property
    def inputs(self):
        """Access the inputs section."""
        return get_simulation_inputs(self.to_libsonata)

    @property
    def run(self):
        """Access to the complete run dictionary for this simulation."""
        return self.to_libsonata.run

    @property
    def time_start(self):
        """Returns the starting time of the simulation."""
        return 0

    @property
    def time_stop(self):
        """Returns the stopping time of the simulation."""
        return self.run.tstop

    @property
    def dt(self):
        """Returns the frequency of reporting in milliseconds."""
        return self.run.dt

    @property
    def time_units(self):
        """Returns the times unit for this simulation."""
        # Assuming ms at the simulation level
        return "ms"

    @property
    def conditions(self):
        """Access to the conditions dictionary for this simulation."""
        return getattr(self.to_libsonata, "conditions", None)

    @property
    def simulator(self):
        """Returns the targeted simulator."""
        target_simulator = getattr(self.to_libsonata, "target_simulator", None)
        return target_simulator.name if target_simulator is not None else None

    @cached_property
    def node_sets(self):
        """Returns the NodeSets object bound to the simulation."""
        try:
            path = self.circuit.to_libsonata.node_sets_path
        except BluepySnapError:  # Error raised if circuit can not be instantiated
            path = ""

        node_sets = NodeSets.from_file(path) if path else NodeSets.from_dict({})

        if self.to_libsonata.node_sets_file != path:
            overwritten = node_sets.update(NodeSets.from_file(self.to_libsonata.node_sets_file))
            _warn_on_overwritten_node_sets(overwritten)

        return node_sets

    @cached_property
    def spikes(self):
        """Access to the SpikeReport."""
        from bluepysnap.spike_report import SpikeReport

        return SpikeReport(self)

    @cached_property
    def reports(self):
        """Access all available FrameReports.

        Notes:
            Supported FrameReports are soma and compartment reports.
        """
        return _collect_frame_reports(self)

    def __getstate__(self):
        """Make Simulations pickle-able, without storing state of caches."""
        return self._simulation_config_path

    def __setstate__(self, state):
        """Load from pickle state."""
        self.__init__(state)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Node population access.

Group Concept
=============

A `Group` is a flexible way to address a series of node IDs, it can take the form of:

- ``int``, ``CircuitNodeId``: return a single node ID if it belongs to the circuit.
- ``CircuitNodeIds`` return IDs of nodes from the node population in an array.
- ``sequence``: return IDs of nodes in an array.
- ``str``: return IDs of nodes in a node set.
- ``mapping``: return IDs of nodes matching a properties filter.
- ``None``: return all node IDs.

If ``group`` is a ``sequence``, the order of results is preserved.
Otherwise the result is sorted and contains no duplicates.

Examples:
    The available group parameter values:

    >>> nodes.ids(group=None)  #  returns all IDs
    >>> nodes.ids(group={})  #  returns all IDs
    >>> nodes.ids(group=1)  #  returns the single ID if present in population
    >>> #  returns the single ID if present in population and the circuit id population
    >>> #  corresponds to nodes.name
    >>> nodes.ids(group=CircuitNodeId('pop', 1))
    >>> nodes.ids(group=[1,2,3])  # returns list of IDs if all present in population
    >>> #  returns list of IDs if all present in population
    >>> nodes.ids(group=CircuitNodeIds.from_dict({"pop": [0, 1,2]}))
    >>> nodes.ids(group="node_set_name")  # returns list of IDs matching node set
    >>> nodes.ids(group={ Node.LAYER: 2})  # returns list of IDs matching layer==2
    >>> nodes.ids(group={ Node.LAYER: [2, 3]})  # returns list of IDs with layer in [2,3]
    >>> nodes.ids(group={ Node.X: (0, 1)})  # returns list of IDs with 0 < x < 1
    >>> # returns list of IDs matching one of the queries inside the 'or' list
    >>> nodes.ids(group={'$or': [{ Node.LAYER: [2, 3]},
    >>>                          { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' }]})
    >>> # returns list of IDs matching all the queries inside the 'and' list
    >>> nodes.ids(group={'$and': [{ Node.LAYER: [2, 3]},
    >>>                           { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' }]})

    See more examples in ``NodePopulation::get``
"""
import inspect
from collections.abc import Mapping, Sequence
from copy import deepcopy

import libsonata
import numpy as np
import pandas as pd
from cached_property import cached_property
from more_itertools import first

from bluepysnap import query, utils
from bluepysnap.circuit_ids import CircuitNodeIds
from bluepysnap.circuit_ids_types import CircuitNodeId
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.node_sets import NodeSet
from bluepysnap.sonata_constants import DYNAMICS_PREFIX, ConstContainer, Node


class NodePopulation:
    """Node population access."""

    def __init__(self, circuit, population_name):
        """Initializes a NodePopulation object.

        Args:
            circuit (bluepysnap.Circuit): the circuit object containing the node population
            population_name (str): the name of the node population

        Returns:
            NodePopulation: A NodePopulation object.
        """
        self._circuit = circuit
        self.name = population_name

    @property
    def _node_sets(self):
        """Node sets defined for this node population."""
        return self._circuit.node_sets

    @cached_property
    def _cache(self):
        """Cached DataFrame of nodes, to be accessed through _get_data()."""
        return pd.DataFrame(index=pd.RangeIndex(self.size, name="node_ids"))

    def _get_libsonata_selection(self, node_ids):
        """Return a libsonata Selection from the given node_ids."""
        if node_ids is None:
            return libsonata.Selection([(0, self.size)])
        return libsonata.Selection(node_ids)

    def _get_values_from_sonata(self, nodes, attr, node_ids):
        """Return the selected values as np.ndarray or pd.Categorical."""
        selection = self._get_libsonata_selection(node_ids)
        if attr in nodes.enumeration_names:
            enumeration = np.asarray(nodes.get_enumeration(attr, selection))
            values = np.asarray(nodes.enumeration_values(attr))
            # if the size of `values` is large enough compared to `enumeration`, not using
            # categorical reduces the memory usage.
            # We compare with nodes.size instead of len(enumeration) to not depend on the selection.
            if len(values) < 0.5 * nodes.size:
                return pd.Categorical.from_codes(enumeration, categories=values)
            return values[enumeration]
        if attr in nodes.attribute_names:
            return nodes.get_attribute(attr, selection)
        if attr.startswith(DYNAMICS_PREFIX):
            stripped = attr[len(DYNAMICS_PREFIX) :]
            if stripped in nodes.dynamics_attribute_names:
                return nodes.get_dynamics_attribute(stripped, selection)
        raise BluepySnapError(f"Attribute not found in population {self.name}: {attr}")

    def _iter_selected_properties(self, existing, desired):
        """Yield ordered (idx, attr) for each attr in desired, and not in existing.

        Called to ensure that the order of the columns of the cached DataFrame doesn't depend
        on the order of the retrieved attributes, when _get_data is called multiple times
        with different properties.

        Args:
            existing: existing attributes, that are going to be skipped.
            desired: desired attributes, that are going to be yielded in order.
        """
        idx = 0
        existing = set(existing)
        desired = set(desired)
        for attr in self._ordered_property_names:
            if attr in existing:
                idx += 1
                continue
            if attr not in desired:
                continue
            yield idx, attr

    def _get_data(self, properties=None, node_ids=None):
        """Collect data for the node population as a pandas.DataFrame.

        Return a DataFrame with node_ids as index, loading the requested properties if needed.

        The returned DataFrame isn't filtered by columns, so it may contain more properties than
        requested, if they were loaded previously.

        This is done for efficiency, since a copy of the data is not needed:
        - in self.get(), the DataFrame is filtered by node_ids first, and by property later
        - in self.property_dtypes(), all the properties are needed
        - in self._node_ids_by_filter(), the required columns are selected if and when needed

        Args:
            properties (str|set|list|None): properties to load, or None to load all of them.
            node_ids (list|np.ndarray|None): node ids to select.
                If None, all the ids are selected, and the cache is read and updated if needed.
                If not None, the cache is read and used if possible, but not updated.
        """
        result = self._cache
        if properties is None:
            properties_set = self.property_names
        else:
            properties_set = set(utils.ensure_list(properties))
            self._check_properties(properties_set)
        if node_ids is not None:
            # Select the ids from the cached dataframe.
            # The original dataframe won't be updated in this case.
            result = result.loc[node_ids]
        cached_columns = properties_set.intersection(result.columns)
        if len(cached_columns) < len(properties_set):
            # some requested properties miss from the cache
            nodes = self._population
            # insert columns at the correct position
            for n, (loc, name) in enumerate(
                self._iter_selected_properties(existing=result.columns, desired=properties_set)
            ):
                values = self._get_values_from_sonata(nodes=nodes, attr=name, node_ids=node_ids)
                result.insert(n + loc, name, values)
        return result

    @cached_property
    def _properties(self):
        """Node population properties."""
        return self._circuit.to_libsonata.node_population_properties(self.name)

    @property
    def _population(self):
        """Libsonata node population.

        Not cached because it would keep the hdf5 file open.
        """
        return self._circuit.to_libsonata.node_population(self.name)

    @cached_property
    def size(self):
        """Node population size."""
        return self._population.size

    @property
    def type(self):
        """Population type."""
        return self._properties.type

    @cached_property
    def _property_names(self):
        return set(self._population.attribute_names)

    @cached_property
    def _dynamics_params_names(self):
        return set(utils.add_dynamic_prefix(self._population.dynamics_attribute_names))

    @cached_property
    def _ordered_property_names(self):
        """Similar to self.property_names, but as an ordered list."""
        return sorted(self._property_names) + sorted(self._dynamics_params_names)

    def source_in_edges(self):
        """Set of edge population names that use this node population as source.

        Returns:
            set: a set containing the names of edge populations using this NodePopulation as
            source.
        """
        return set(
            edge.name for edge in self._circuit.edges.values() if self.name == edge.source.name
        )

    def target_in_edges(self):
        """Set of edge population names that use this node population as target.

        Returns:
            set: a set containing the names of edge populations using this NodePopulation as
            target.
        """
        return set(
            edge.name for edge in self._circuit.edges.values() if self.name == edge.target.name
        )

    @property
    def config(self):
        """Access the configuration for the population.

        This configuration is extended with:

        * 'components' of the circuit config
        * 'nodes_file': the path the h5 file containing the population.
        """
        return self._circuit.get_node_population_config(self.name)

    @property
    def property_names(self):
        """Set of available node properties.

        Notes:
            Properties are a combination of the group attributes and the dynamics_params.
        """
        return self._property_names | self._dynamics_params_names

    def container_property_names(self, container):
        """Lists the ConstContainer properties shared with the NodePopulation.

        Args:
            container (ConstContainer): a container class for node properties.

        Returns:
            list: A list of strings corresponding to the properties that you can use from the
            container class

        Examples:
            >>> from bluepysnap.sonata_constants import Node
            >>> print(my_node_population.container_property_names(Node))
            >>> ["X", "Y", "Z"] # values from Node that you can use with my_node_population
            >>> my_node_population.property_values(Node.X)
            >>> my_node_population.property_values(Node.get("X"))
        """
        if not inspect.isclass(container) or not issubclass(container, ConstContainer):
            raise BluepySnapError("'container' must be a subclass of ConstContainer")
        in_file = self.property_names
        return [k for k in container.key_set() if container.get(k) in in_file]

    def property_values(self, prop, is_present=False):
        """Set of values for a given property.

        Args:
           prop (str): Name of the property to retrieve.
           is_present (bool): if the field is categorical it forces the values to be actually
           present inside the dataset. (see: Notes for more information).

        Returns:
            set: A set of the unique values of the property in the node population.

        Notes:
            For categorical fields returning 'unique' values can be confusing, see:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/categorical.html
            in #working-with-categories.
            The is_present argument forces the unique() even on the categorical fields.
        """
        res = self.get(properties=prop)
        if isinstance(res.dtype, pd.CategoricalDtype) and not is_present:
            return set(res.cat.categories)
        return set(res.unique())

    @cached_property
    def property_dtypes(self):
        """Returns the dtypes of all the properties.

        Returns:
            pandas.Series: series indexed by field name with the corresponding dtype as value.
        """
        # read all the properties, without loading any node id
        return self._get_data(properties=None, node_ids=[]).dtypes.sort_index()

    def _check_id(self, node_id):
        """Check that single node ID belongs to the circuit."""
        if node_id < 0 or node_id >= self.size:
            raise BluepySnapError(f"node ID not found: {node_id} in population '{self.name}'")

    def _check_ids(self, node_ids):
        """Check that node IDs belong to the circuit."""
        if len(node_ids) == 0:
            return
        # use the function with better performance for arrays or lists
        if isinstance(node_ids, np.ndarray):
            max_id = node_ids.max()
            min_id = node_ids.min()
        else:
            max_id = max(node_ids)
            min_id = min(node_ids)
        if min_id < 0 or max_id >= self.size:
            raise BluepySnapError(
                f"All node IDs must be >= 0 and < {self.size} " f"for population '{self.name}'"
            )

    def _check_properties(self, properties):
        """Check if the properties exist inside the dataset."""
        unknown_props = properties - self.property_names
        if unknown_props:
            raise BluepySnapError(f"Unknown node properties: {sorted(unknown_props)}")

    def _resolve_nodesets(self, queries, raise_missing_prop):
        def _resolve(queries, queries_key):
            if queries_key == query.NODE_SET_KEY:
                if query.AND_KEY not in queries:
                    queries[query.AND_KEY] = []
                node_set = self._node_sets[queries[queries_key]]
                queries[query.AND_KEY].append(
                    {query.NODE_ID_KEY: node_set.get_ids(self._population, raise_missing_prop)}
                )
                del queries[queries_key]

        resolved_queries = deepcopy(queries)
        query.traverse_queries_bottom_up(resolved_queries, _resolve)
        return resolved_queries

    def _node_ids_by_filter(self, queries, raise_missing_prop):
        """Return node IDs if their properties match the `queries` dict.

        `props` values could be:
            pairs (range match for floating dtype fields)
            scalar or iterables (exact or "one of" match for other fields)

        You can use the special operators '$or' and '$and' also to combine different queries
        together.

        Examples:
            >>> _node_ids_by_filter({ Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' })
            >>> _node_ids_by_filter({ Node.LAYER: [2, 3] })
            >>> _node_ids_by_filter({'$or': [{ Node.LAYER: [2, 3]},
            >>>                              { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' }]})

        """
        queries = self._resolve_nodesets(queries, raise_missing_prop)
        properties = query.get_properties(queries)
        if raise_missing_prop:
            self._check_properties(properties)
        # load all the properties needed to execute the query, excluding the unknown properties
        data = self._get_data(properties & self.property_names)
        idx = query.resolve_ids(data, self.name, queries)
        return idx.nonzero()[0]

    def ids(self, group=None, limit=None, sample=None, raise_missing_property=True):
        """Node IDs corresponding to node ``group``.

        Args:
            group (int/CircuitNodeId/CircuitNodeIds/sequence/str/mapping/None):
                See :ref:`Group Concept`
            sample (int): If specified, randomly choose ``sample`` number of
                IDs from the match result. If the size of the sample is greater than
                the size of the NodePopulation then all ids are taken and shuffled.

            limit (int): If specified, return the first ``limit`` number of
                IDs from the match result. If limit is greater than the size of the population
                all node IDs are returned.

            raise_missing_property (bool): if True, raises if a property is not listed in this
                population. Otherwise the ids are just not selected if a property is missing.

        Returns:
            numpy.array: A numpy array of IDs.
        """
        # pylint: disable=too-many-branches
        preserve_order = False
        if isinstance(group, str):
            group = self._node_sets[group]
        elif isinstance(group, CircuitNodeIds):
            group = group.filter_population(self.name).get_ids()

        if group is None:
            result = np.arange(self.size)
        elif isinstance(group, NodeSet):
            result = group.get_ids(self._population, raise_missing_property)
        elif isinstance(group, Mapping):
            result = self._node_ids_by_filter(group, raise_missing_property)
        elif isinstance(group, np.ndarray):
            result = group
            self._check_ids(result)
            preserve_order = True
        else:
            result = utils.ensure_list(group)
            # test if first value is a CircuitNodeId all values are all CircuitNodeId
            if isinstance(first(result, None), CircuitNodeId):
                try:
                    result = [cid.id for cid in result if cid.population == self.name]
                except AttributeError as e:
                    raise BluepySnapError(
                        "All values from a list must be of type int or CircuitNodeId."
                    ) from e
            self._check_ids(result)
            preserve_order = isinstance(group, Sequence)

        if sample is not None:
            if len(result) > 0:
                result = np.random.choice(result, min(sample, len(result)), replace=False)
            preserve_order = False
        if limit is not None:
            result = result[:limit]

        result = utils.ensure_ids(result)
        if preserve_order:
            return result
        else:
            return np.unique(result)

    def get(self, group=None, properties=None, raise_missing_property=True):
        """Node properties as a pandas Series or DataFrame.

        Args:
            group (int/CircuitNodeId/CircuitNodeIds/sequence/str/mapping/None):
                see :ref:`Group Concept`
            properties (list|str|None): If specified, return only the properties in the list.
                Otherwise, return all the properties.
            raise_missing_property (bool): when a property used for filtering ids is not present in
                the population, raise an error if True, or return empty Series/DataFrame if False.

        Returns:
            value/pandas.Series/pandas.DataFrame:
                The type of the returned object depends on the type of the input parameters,
                see the Examples for an explanation of the different cases.

        Notes:
            The NodePopulation.property_names function will give you all the usable properties
            for the `properties` argument.

        Examples:
            Considering a node population composed by 3 nodes (0, 1, 2) and 12 properties,
            the following examples show the types of the returned objects.

            - If ``group`` is a single node ID and ``properties`` a single property,
              returns a single scalar value.

                >>> result = my_node_population.get(group=0, properties=Cell.MTYPE)
                >>> type(result)
                str

            - If ``group`` is a single node ID and ``properties`` a list or None,
              returns a pandas Series indexed by the properties.

                >>> result = my_node_population.get(group=0)
                >>> type(result), result.shape
                (pandas.core.series.Series, (12,))

                >>> result = my_node_population.get(group=0, properties=[Cell.MTYPE])
                >>> type(result), result.shape
                (pandas.core.series.Series, (1,))

            - If ``group`` is anything other than a single node ID, and ``properties`` is a single
              property, returns a pandas Series indexed by node IDs.

                >>> result = my_node_population.get(properties=Cell.MTYPE)
                >>> type(result), result.shape
                (pandas.core.series.Series, (3,))

                >>> result = my_node_population.get(group=[0], properties=Cell.MTYPE)
                >>> type(result), result.shape
                (pandas.core.series.Series, (1,))

            - In all the other cases, returns a pandas DataFrame indexed by node IDs.

                >>> result = my_node_population.get()
                >>> type(result), result.shape
                (pandas.core.frame.DataFrame, (3, 12))

                >>> result = my_node_population.get(group=[0])
                >>> type(result), result.shape
                (pandas.core.frame.DataFrame, (1, 12))

                >>> result = my_node_population.get(properties=[Cell.MTYPE])
                >>> type(result), result.shape
                (pandas.core.frame.DataFrame, (3, 1))

                >>> result = my_node_population.get(group=[0], properties=[Cell.MTYPE])
                >>> type(result), result.shape
                (pandas.core.frame.DataFrame, (1, 1))
        """
        if group is not None:
            if isinstance(group, (int, np.integer)):
                self._check_id(group)
            elif isinstance(group, CircuitNodeId):
                group = self.ids(group, raise_missing_property=raise_missing_property)[0]
            else:
                group = self.ids(group, raise_missing_property=raise_missing_property)

        result = self._get_data(properties=properties)
        result = result.loc[group] if group is not None else result
        result = result[properties] if properties is not None else result
        return result

    def positions(self, group=None):
        """Node position(s) as pandas Series or DataFrame.

        Args:
            group (int/CircuitNodeId/CircuitNodeIds/sequence/str/mapping/None): Which nodes will
                have their positions returned depends on the type of the ``group`` argument:

                - ``int``, ``CircuitNodeId``: return the position of a single node.
                - ``CircuitNodeIds`` return the position from a NodeCircuitNodeIds.
                - ``sequence``: return the positions from a list of node IDs.
                - ``str``: return the positions of nodes in a node set.
                - ``mapping``: return the positions of nodes matching a properties filter.
                - ``None``: return the positions of all nodes.

        Returns:
            pandas.Series/pandas.DataFrame:
                Series of ('x', 'y', 'z') if single node ID is

                passed as ``group``. Otherwise, a pandas.DataFrame of ('x', 'y', 'z') indexed
                by node IDs.
        """
        result = self.get(group=group, properties=[Node.X, Node.Y, Node.Z])
        return result.astype(float)

    def orientations(self, group=None):
        """Node orientation(s) as a pandas numpy array or pandas Series.

        Args:
            group (int/CircuitNodeId/CircuitNodeIds/sequence/str/mapping/None): Which nodes will
                have their positions returned depends on the type of the ``group`` argument:

                - ``int``, ``CircuitNodeId``: return the orientation of a single node.
                - ``CircuitNodeIds`` return the orientation from a NodeCircuitNodeIds.
                - ``sequence``: return the orientations from a list of node IDs.
                - ``str``: return the orientations of nodes in a node set.
                - ``mapping``: return the orientations of nodes matching a properties filter.
                - ``None``: return the orientations of all nodes.

        Returns:
            numpy.ndarray/pandas.Series:
                A 3x3 rotation matrix if a single node ID is passed as ``group``.
                Otherwise a pandas Series with rotation matrices indexed by node IDs.
        """
        # need to keep this quaternion ordering for quaternion2mat (expects w, x, y , z)
        props = np.array(
            [Node.ORIENTATION_W, Node.ORIENTATION_X, Node.ORIENTATION_Y, Node.ORIENTATION_Z]
        )
        props_mask = np.isin(props, list(self.property_names))
        orientation_count = np.count_nonzero(props_mask)
        if orientation_count == 4:
            trans = utils.quaternion2mat
        elif orientation_count in [1, 2, 3]:
            raise BluepySnapError(
                "Missing orientation fields. Should be 4 quaternions or euler angles or nothing"
            )
        else:
            # need to keep this rotation_angle ordering for euler2mat (expects z, y, x)
            props = np.array(
                [
                    Node.ROTATION_ANGLE_Z,
                    Node.ROTATION_ANGLE_Y,
                    Node.ROTATION_ANGLE_X,
                ]
            )
            props_mask = np.isin(props, list(self.property_names))
            trans = utils.euler2mat

        result = self.get(group=group, properties=props[props_mask])

        def _get_values(prop):
            """Retrieve prop from the result Dataframe/Series."""
            if isinstance(result, pd.Series):
                return [result.get(prop, 0)]
            return result.get(prop, np.zeros((result.shape[0],)))

        args = [_get_values(prop) for prop in props]
        if utils.is_node_id(group):
            return trans(*args)[0]
        return pd.Series(trans(*args), index=result.index, name="orientation")

    def count(self, group=None):
        """Total number of nodes for a given node group.

        Args:
            group (int/CircuitNodeId/CircuitNodeIds/sequence/str/mapping/None): Which nodes will
                have their positions returned depends on the type of the ``group`` argument:

                - ``int``, ``CircuitNodeId``: return the count of a single node.
                - ``CircuitNodeIds`` return the count of nodes from a NodeCircuitNodeIds.
                - ``sequence``: return the count of nodes from a list of node IDs.
                - ``str``: return the count of nodes in a node set.
                - ``mapping``: return count of nodes matching a properties filter.
                - ``None``: return the count of all nodes.

        Returns:
            int: The total number of nodes in a given group.
        """
        return len(self.ids(group))

    @cached_property
    def morph(self):
        """Access to node morphologies."""
        from bluepysnap.morph import MorphHelper

        return MorphHelper(
            self._properties.morphologies_dir,
            self,
            alternate_morphologies=self._properties.alternate_morphology_formats,
        )

    @cached_property
    def models(self):
        """Access to node neuron models."""
        from bluepysnap.neuron_models import NeuronModelsHelper

        return NeuronModelsHelper(self._properties, self)

    @property
    def h5_filepath(self):
        """Get the H5 nodes file associated with population."""
        return self.config["nodes_file"]

    @cached_property
    def spatial_segment_index(self):
        """Access to edges spatial index."""
        try:
            from spatial_index import open_index
        except ImportError as e:
            raise BluepySnapError(
                (
                    "Spatial index is for now only available internally to BBP. ",
                    "It requires `spatial_index`, an internal package.",
                )
            ) from e

        index_dir = self._properties.spatial_segment_index_dir
        if not index_dir:
            raise BluepySnapError(f"It appears {self.name} does not have segment indices")
        return open_index(index_dir)

    def __getstate__(self):
        """Make NodePopulation pickle-able, without storing state of caches."""
        return (self._circuit, self.name)

    def __setstate__(self, state):
        """Load from pickle state."""
        circuit, name = state
        self.__init__(circuit=circuit, population_name=name)



import json
import logging
import tempfile

import pytest

import bluepysnap.circuit as test_module
from bluepysnap.exceptions import BluepySnapError
from bluepysnap.sonata_constants import Edge, Node

from utils import TEST_DATA_DIR


def test_partial_circuit_config_minimal():
    config = {
        "metadata": {"status": "partial"},
        "networks": {
            "nodes": [
                {
                    # missing `nodes_file` rises libsonata exception
                    "nodes_file": str(TEST_DATA_DIR / "nodes.h5"),
                    "populations": {
                        "default": {},
                    },
                }
            ],
            "edges": [
                {
                    # missing `edges_file` rises libsonata exception
                    "edges_file": str(TEST_DATA_DIR / "edges.h5"),
                    "populations": {
                        "default": {},
                    },
                }
            ],
        },
    }
    with tempfile.NamedTemporaryFile(mode="w+") as config_file:
        config_file.write(json.dumps(config))
        config_file.flush()
        circuit = test_module.Circuit(config_file.name)

    assert circuit.get_node_population_config("default")
    assert circuit.nodes
    assert circuit.nodes["default"].type == "biophysical"
    assert circuit.nodes["default"].size == 3
    assert circuit.nodes["default"].source_in_edges() == {"default"}
    assert circuit.nodes["default"].target_in_edges() == {"default"}
    assert circuit.nodes["default"].config is not None
    assert circuit.nodes["default"].property_names is not None
    assert circuit.nodes["default"].container_property_names(Node) is not None
    assert circuit.nodes["default"].container_property_names(Edge) == []
    assert circuit.nodes["default"].property_values("layer") == {2, 6}
    assert circuit.nodes["default"].property_dtypes is not None
    assert list(circuit.nodes["default"].ids()) == [0, 1, 2]
    assert circuit.nodes["default"].get() is not None
    assert circuit.nodes["default"].positions() is not None
    assert circuit.nodes["default"].orientations() is not None
    assert circuit.nodes["default"].count() == 3
    assert circuit.nodes["default"].morph is not None
    assert circuit.nodes["default"].models is not None
    assert circuit.nodes["default"].h5_filepath == str(TEST_DATA_DIR / "nodes.h5")
    assert circuit.nodes["default"]._properties.spatial_segment_index_dir == ""

    assert circuit.nodes.population_names == ["default"]
    assert list(circuit.nodes.values())

    assert [item.id for item in circuit.nodes.ids()] == [0, 1, 2]

    assert circuit.get_edge_population_config("default")
    assert circuit.edges
    assert circuit.edges["default"].type == "chemical"
    assert circuit.edges["default"].size == 4
    assert circuit.edges["default"].source is not None
    assert circuit.edges["default"].target is not None
    assert circuit.edges["default"].config is not None
    assert circuit.edges["default"].property_names is not None
    assert circuit.edges["default"].property_dtypes is not None
    assert circuit.edges["default"].container_property_names(Node) == []
    assert circuit.edges["default"].container_property_names(Edge) is not None
    assert list(circuit.edges["default"].ids()) == [0, 1, 2, 3]
    assert list(circuit.edges["default"].get([1, 2], None)) == [1, 2]
    assert circuit.edges["default"].positions([1, 2], "afferent", "center") is not None
    assert list(circuit.edges["default"].afferent_nodes(None)) == [0, 2]
    assert list(circuit.edges["default"].efferent_nodes(None)) == [0, 1]
    assert list(circuit.edges["default"].pathway_edges(0)) == [1, 2]
    assert list(circuit.edges["default"].afferent_edges(0)) == [0]
    assert list(circuit.edges["default"].efferent_edges(0)) == [1, 2]
    assert circuit.edges["default"].iter_connections() is not None
    assert circuit.edges["default"].h5_filepath == str(TEST_DATA_DIR / "edges.h5")
    assert circuit.nodes["default"]._properties.spatial_segment_index_dir == ""

    assert circuit.edges.population_names == ["default"]
    assert list(circuit.edges.values())

    assert [item.id for item in circuit.edges.ids()] == [0, 1, 2, 3]


def test_partial_circuit_config_log(caplog):
    caplog.set_level(logging.INFO)

    config = {"metadata": {"status": "partial"}}

    with tempfile.NamedTemporaryFile(mode="w+") as config_file:
        config_file.write(json.dumps(config))
        config_file.flush()
        test_module.Circuit(config_file.name)

    assert "Loaded PARTIAL circuit config" in caplog.text


def test_partial_circuit_config_empty():
    config = {"metadata": {"status": "partial"}}
    with tempfile.NamedTemporaryFile(mode="w+") as config_file:
        config_file.write(json.dumps(config))
        config_file.flush()
        circuit = test_module.Circuit(config_file.name)

    with pytest.raises(BluepySnapError):
        assert circuit.get_node_population_config("default")
    with pytest.raises(BluepySnapError):
        assert circuit.get_edge_population_config("default")
    with pytest.raises(BluepySnapError):
        circuit.nodes.ids()
    with pytest.raises(BluepySnapError):
        circuit.edges.ids()
